# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from copy import deepcopy
import math
from typing import List, Optional

import numpy as np
from packaging import version
import torch
from torch import Tensor
import torch.nn as nn

try:
    from torch.linalg import LinAlgError
except:
    LinAlgError = RuntimeError
import warnings

import unfoldNd

from brevitas import torch_version
from brevitas.graph.calibrate import disable_return_quant_tensor
from brevitas.graph.calibrate import restore_return_quant_tensor
from brevitas.graph.gpxq import GPxQ
from brevitas.graph.gpxq import gpxq_mode
from brevitas.graph.gpxq import StopFwdException
from brevitas.graph.gpxq import SUPPORTED_CONV_OP
from brevitas.graph.gpxq import SUPPORTED_TCONV_OP
import brevitas.nn as qnn
from brevitas.quant_tensor import _unpack_quant_tensor


class gpfq_mode(gpxq_mode):
    """
    Apply GPFQ algorithm.

    Args:
        model (Module): The model to quantize with GPFQ
        group_of_parallel_layers (Optional, List[str]): .List of lists where each inner list is a group
            of layer names that can be optimized in parallel. Default: None
        inplace (bool): Wheter to apply GPFQ inplace or perform a deepcopy. Default: True
        create_weight_orig (bool): If True, store the original floating point weights before applying
            gpfq. These weights will be used anytime quantization is disabled. Default: True
        use_quant_activations (bool): Wheter to leave quantize activations enabled while performing
            GPFQ. Default: False
        p (float): The percentage of processed inputs to use. Default: 1.0
        return_forward_output (bool): If True, returns the output of the forward pass. Otherwise the
            forward call inside the context manager returns None. Default: False
        act_order (bool): Whether to order greedy path following by Hessian approximation. Default: False

    Example:
        >>> with torch.no_grad():
        >>>     with gpfq_mode(model) as gpfq:
        >>>         gpfq_model = gpfq.model
        >>>         for i in tqdm(range(gpfq.num_layers)):
        >>>             for img, t in calib_loader:
        >>>                 img = img.cuda()
        >>>                 gpfq_model(img)
        >>>             gpfq.update()
    """

    def __init__(
            self,
            model: nn.Module,
            group_of_parallel_layers: Optional[List[str]] = None,
            inplace: bool = True,
            create_weight_orig: bool = True,
            use_quant_activations: bool = True,
            p: float = 1.0,
            return_forward_output: bool = False,
            act_order: bool = False,
            gpfq_class: Optional[nn.Module] = None) -> None:
        if not inplace:
            model = deepcopy(model)
        super().__init__(
            model,
            group_of_parallel_layers,
            inplace,
            create_weight_orig,
            use_quant_activations,
            act_order,
            return_forward_output)

        self.p = p
        if gpfq_class is None:
            gpfq_class = GPFQ
        self.gpfq_class = gpfq_class
        assert issubclass(gpfq_class, GPxQ), \
            "Error: expected `gpfq_class` to be derived from `brevitas.graph.gpxq.GPxQ`."

    def catch_stopfwd(self, *args, **kwargs):
        # Collect quant input
        try:
            self.orig_forward(*args, **kwargs)
        except StopFwdException:
            pass

        # Disable quantization
        self.return_quant_tensor_state = disable_return_quant_tensor(self.model)
        self.disable_quant_inference.disable_param_quantization(self.model, is_training=False)
        self.disable_quant_inference.disable_act_quantization(self.model, is_training=False)
        # Collect float input
        try:
            self.orig_forward(*args, **kwargs)
        except StopFwdException:
            pass

        # Re-enable quantization. If activation quantization is disabled,
        # we also disable bias quantization
        self.disable_quant_inference.enable_param_quantization(self.model, is_training=False)
        if self.use_quant_activations:
            self.disable_quant_inference.enable_act_quantization(self.model, is_training=False)
        else:
            self.disable_quant_inference.disable_bias_quantization(self.model, is_training=False)
        restore_return_quant_tensor(self.model, self.return_quant_tensor_state)

        if self.return_forward_output:
            # If we want to return the output of the network, we need to disable all hooks
            for name, gpxq_class in self.gpxq_layers.items():
                gpxq_class.disable_pre_forward_hook = True
            out = self.orig_forward(*args, **kwargs)
            for name, gpxq_class in self.gpxq_layers.items():
                gpxq_class.disable_pre_forward_hook = False
            return out

    def initialize_module_optimizer(
            self, layer, name, act_order, len_parallel_layers, create_weight_orig):
        return self.gpfq_class(
            layer=layer,
            name=name,
            act_order=act_order,
            len_parallel_layers=len_parallel_layers,
            create_weight_orig=create_weight_orig,
            p=self.p)


class GPFQ(GPxQ):
    """
    Based on https://github.com/YixuanSeanZhou/Quantized_Neural_Nets/tree/main
    """

    def __init__(self, layer, name, act_order, len_parallel_layers, create_weight_orig, p) -> None:

        super().__init__(layer, name, act_order, len_parallel_layers, create_weight_orig)

        self.float_input = None
        self.quant_input = None
        self.index_computed = False
        self.p = p

    def update_batch(self, module, input, current_layer):
        if self.disable_pre_forward_hook:
            return input

        # Update reference to current layer
        current_layer.layer_names.add(self.name)
        is_quant_enabled = module.weight_quant.is_quant_enabled

        inp = self.process_input(input)
        batch_size = inp.shape[0]

        # Preprocess the input to compute the Hessian
        if isinstance(self.layer, qnn.QuantLinear):
            if len(inp.shape) > 2:
                inp = inp.reshape((-1, sum(inp.shape[2:])))
            # For QuantLinear layer, groups will be 1
            inp_processed = inp.unsqueeze(0)

        if isinstance(self.layer, SUPPORTED_CONV_OP):
            # Pick the correct unfoldNd class
            if isinstance(self.layer, SUPPORTED_TCONV_OP):
                unfold_impl = unfoldNd.UnfoldTransposeNd
            else:
                unfold_impl = unfoldNd.UnfoldNd

            unfold = unfold_impl(
                self.layer.kernel_size,
                dilation=self.layer.dilation,
                padding=self.layer.padding,
                stride=self.layer.kernel_size)

            # Split input based on how many groups in convolution
            inp_by_group = torch.chunk(inp, self.groups, 1)
            inp_processed = []
            # Preprocess input by group
            for i, inp in enumerate(inp_by_group):

                inp = unfold(inp)

                batch_size, num_blocks = inp.shape[0], inp.shape[-1]
                inp = torch.transpose(inp, 1, 2)  # shape (B, L, C*kernel_size[0]*kernel_size[1])
                inp = inp.reshape(-1, inp.size(-1))  # shape (B*L, C*kernel_size[0]*kernel_size[1])

                if not self.index_computed:
                    self.index_computed = True
                    self.rand_indices = np.concatenate([
                        np.random.choice(
                            np.arange(num_blocks * i, num_blocks * (i + 1)),
                            size=int(
                                self.p * num_blocks + 1 if self.p != 1 else self.p * num_blocks))
                        for i in range(batch_size)])  # need to define self.p (probability)

                indexes = self.rand_indices
                if np.max(self.rand_indices) > inp.shape[0]:
                    indexes = self.rand_indices < inp.shape[0]
                    indexes = self.rand_indices[indexes]

                inp = inp[indexes]
                inp_processed.append(inp)
            inp_processed = torch.stack(inp_processed)

        if not is_quant_enabled:
            if self.float_input is None:
                self.float_input = inp_processed
            else:
                self.float_input = torch.cat([self.float_input, inp_processed], dim=1)
        else:
            if self.quant_input is None:
                self.quant_input = inp_processed
            else:
                self.quant_input = torch.cat([self.quant_input, inp_processed], dim=1)
        # If we are executing GPFQ with group of parallel layers, we keep track of how many forward
        # we executed. Once we executed as many as the number of parallel_layers, we raise
        # StopFwdException
        current_layer.forward_count += 1
        if current_layer.forward_count == self.len_parallel_layers:
            current_layer.forward_count = 0
            raise StopFwdException

    def single_layer_update(self):
        assert not self.layer.weight_quant.requires_quant_input, \
            "Error: GPFQ does not support weight quantizers that require quantized inputs."
        weight = self.layer.weight.data
        dev = weight.device
        dtype = weight.dtype
        if isinstance(self.layer, SUPPORTED_CONV_OP):
            if isinstance(self.layer, SUPPORTED_TCONV_OP):
                weight = weight.transpose(1, 0)  # This performs a view
            weight = weight.flatten(1)
        weight = weight.view(self.groups, -1, weight.shape[-1])  # [Groups, OC/Groups, IC]

        self.float_input = self.float_input.to(dev)
        self.quant_input = self.quant_input.to(dev)
        U = torch.zeros(
            weight.shape[0], weight.shape[1], self.float_input.shape[1], device=dev, dtype=dtype)
        # We don't need full Hessian, we just need the diagonal
        # Summing over batch dimension
        H_diag = self.quant_input.transpose(2, 1).square().sum(2)
        permutation_list = []
        for group_index in range(self.groups):
            if self.act_order:
                # Re-order Hessian_diagonal so that weights associated to
                # higher magnitude activations are quantized first
                perm = torch.argsort(H_diag[group_index, :], descending=True)
            else:
                # No permutation, permutation tensor is a ordered index
                perm = torch.tensor(range(weight.shape[-1]), device=dev)
            permutation_list.append(perm)
        del H_diag
        for t in range(weight.shape[-1]):
            for group_index in range(self.groups):
                U[group_index] += torch.matmul(
                    weight[group_index, :, permutation_list[group_index][t]].unsqueeze(1),
                    self.float_input[group_index, :, permutation_list[group_index][t]].unsqueeze(
                        0))  #[OC/Groups, 1] * [1, INSHAPE[1]]
                norm = torch.linalg.norm(
                    self.quant_input[group_index, :, permutation_list[group_index][t]], 2) ** 2
                if norm > 0:
                    q_arg = U[group_index].matmul(
                        self.quant_input[group_index, :, permutation_list[group_index][t]]) / norm
                else:
                    q_arg = torch.zeros_like(U[group_index, :, 0])

                weight[group_index, :, permutation_list[group_index][t]] = q_arg
            q = self.get_quant_weights(t, 0, permutation_list)
            for group_index in range(self.groups):
                U[group_index] -= torch.matmul(
                    q[group_index].unsqueeze(1),
                    self.quant_input[group_index, :, permutation_list[group_index][t]].unsqueeze(0))

        del self.float_input
        del self.quant_input


class GPFQv2(GPFQ):
    """
    Memory-efficient GPFQ formulation introduced in https://arxiv.org/pdf/2409.17092
    """

    def __init__(self, layer, name, act_order, len_parallel_layers, create_weight_orig, p) -> None:
        super().__init__(layer, name, act_order, len_parallel_layers, create_weight_orig, p)
        # Initialize covariance matrices. We need it in float32 to compute the inverse
        # H = (\hat{X} \hat{X}^T)^{1/2}
        self.H: Tensor = torch.zeros((self.groups, self.columns, self.columns),
                                     device="cpu",
                                     dtype=torch.float32)
        # G = X \hat{X}^T
        self.G: Tensor = torch.zeros((self.groups, self.columns, self.columns),
                                     device="cpu",
                                     dtype=torch.float32)
        # buffer to speed-up GPU to CPU transfer
        self.B: Tensor = torch.zeros((self.groups, self.columns, self.columns),
                                     device="cpu",
                                     dtype=torch.float32,
                                     pin_memory=torch.cuda.is_available())
        self.nsamples = 0

        assert torch_version >= version.parse('1.10'), "GPFQv2 requires torch 1.10 or higher"

    def update_batch(self, module, input, current_layer):
        if self.disable_pre_forward_hook:
            return input

        # Update reference to current layer
        current_layer.layer_names.add(self.name)
        is_quant_enabled = module.weight_quant.is_quant_enabled

        inp = self.process_input(input)
        inp = _unpack_quant_tensor(inp)

        # Preprocess the input to compute the Hessian
        if isinstance(self.layer, qnn.QuantLinear):
            if len(inp.shape) > 2:
                inp = inp.reshape((-1, sum(inp.shape[2:])))
            inp = inp.t()
            # For QuantLinear layer, groups will be 1
            inp_processed = inp.unsqueeze(0)

        if isinstance(self.layer, SUPPORTED_CONV_OP):
            # Pick the correct unfoldNd class
            if isinstance(self.layer, SUPPORTED_TCONV_OP):
                unfold_impl = unfoldNd.UnfoldTransposeNd
            else:
                unfold_impl = unfoldNd.UnfoldNd

            unfold = unfold_impl(
                self.layer.kernel_size,
                dilation=self.layer.dilation,
                padding=self.layer.padding,
                stride=self.layer.stride)

            # Split input based on how many groups in convolution
            inp_by_group = torch.chunk(inp, self.groups, 1)
            inp_processed = []
            # Preprocess input by group
            for inp in inp_by_group:
                inp = unfold(inp)
                inp = inp.transpose(1, 0)
                inp = inp.flatten(1)
                inp_processed.append(inp)
            inp_processed = torch.stack(inp_processed)

        # NOTE: in the gpfq_mode context manager, we first collect quant inputs, then
        # we collect float inputs for the same batch. We assume this pattern here, but
        # will add a check just in case.
        n = inp_processed.size(1)
        inp_processed = math.sqrt(2 / n) * inp_processed.to(torch.float32)

        # if quant is not enabled, then it is the float input; if it is a float input
        # then a quant input has already happened and we can update G
        if not is_quant_enabled:
            # Computing the normalized H matrix using CPU buffer
            self.B.copy_(self.quant_input.bmm(inp_processed.transpose(2, 1)))
            self.G += self.B
            self.quant_input = None  # NOTE: set back to None now that we've used it
        else:
            # Computing the normalized H matrix using CPU buffer
            self.B.copy_(inp_processed.bmm(inp_processed.transpose(2, 1)))
            self.H += self.B
            # store the quantized input for computing the H matrix
            assert self.quant_input is None
            self.quant_input = inp_processed

        # If we are executing GPFQ with group of parallel layers, we keep track of how many forward
        # we executed. Once we executed as many as the number of parallel_layers, we raise
        # StopFwdException
        current_layer.forward_count += 1
        if current_layer.forward_count == self.len_parallel_layers:
            current_layer.forward_count = 0
            raise StopFwdException

    def _get_permutation_list(self, weight: Tensor):
        permutation_list = []
        if self.act_order:
            # We don't need full Hessian, we just need the diagonal
            H_diag = self.quant_input.transpose(2, 1).square().sum(2)
            for group_index in range(self.groups):
                # Re-order Hessian_diagonal so that weights associated to
                # higher magnitude activations are quantized first
                perm = torch.argsort(H_diag[group_index, :], descending=True)
                perm = perm.to(weight.device)
                permutation_list.append(perm)
        else:
            for group_index in range(self.groups):
                # No permutation, permutation tensor is a ordered index
                perm = torch.tensor(range(weight.shape[-1]), device=weight.device)
                permutation_list.append(perm)
        return permutation_list

    def single_layer_update(self, percdamp: float = 0.01):
        assert not self.layer.weight_quant.requires_quant_input, \
            "Error: GPFQ does not support weight quantizers that require quantized inputs."
        weight = self.layer.weight.data
        dev = weight.device
        dtype = weight.dtype
        if isinstance(self.layer, SUPPORTED_CONV_OP):
            if isinstance(self.layer, SUPPORTED_TCONV_OP):
                weight = weight.transpose(1, 0)  # This performs a view
            weight = weight.flatten(1)
        weight = weight.view(self.groups, -1, weight.shape[-1])  # [Groups, OC/Groups, IC]

        # stablize H with a dampening factor and then square root the matrix
        norms = torch.zeros((self.groups, self.columns), device=dev, dtype=dtype)
        self.H = self.H.to(dev)
        diag = torch.arange(self.columns, device='cpu')
        for i in range(self.groups):
            damp = percdamp * self.H[i].diag().mean()
            self.H[i, diag, diag] += damp
            norms[i] = self.H[i].diag()  # set the norms post-dampening
            eigvals, eigvecs = torch.linalg.eigh(self.H[i])
            eigvals.clamp_min_(0.0).sqrt_()  # should be positive-definite
            self.H[i] = eigvecs @ torch.diag(eigvals) @ eigvecs.t()
        del eigvecs, eigvals, diag
        self.quant_input = self.H  # NOTE: do this here for the `get_permutation_list` function

        # Try/Except in case the inverse of H cannot be computed
        try:
            self.float_input = self.H.clone()  # going to calculate H^{-1} here
            for i in range(self.groups):
                # from our matrix sqrt, we know G is symmetric and positive-definite, so we
                # can use Cholesky decomposition as an efficient, numerically stable inverse
                L = torch.linalg.cholesky(self.float_input[i])
                self.float_input[i] = torch.cholesky_inverse(L)
            self.float_input = torch.bmm(self.float_input.to(dev), self.G.to(dev))
            del L  # memory management
        except LinAlgError:
            warnings.warn(
                f'Failed to compute the inverse of H for layer {self.name} '
                f'GPFQ will not be applied. '
                f'Increasing the number of samples might fix this issue')
            return
        finally:
            del self.H, self.G, self.B  # memory management

        permutation_list = self._get_permutation_list(weight)

        U = torch.zeros(
            weight.shape[0], weight.shape[1], self.float_input.shape[1], device=dev,
            dtype=dtype)  # [Groups, OC/groups, Samples]

        for t in range(weight.shape[-1]):
            for group_index in range(self.groups):
                i = permutation_list[group_index][t]
                U[group_index] += torch.matmul(
                    weight[group_index, :, i].unsqueeze(1),
                    self.float_input[group_index, :, i].unsqueeze(0),
                )  # [OC/Groups, 1] * [1, INSHAPE[1]]
                norm = norms[group_index, i]
                if norm > 0:
                    q_arg = U[group_index].matmul(self.quant_input[group_index, :, i]) / norm
                else:
                    q_arg = torch.zeros_like(U[group_index, :, 0])
                weight[group_index, :, i] = q_arg
            q_groups = self.get_quant_weights(t, 0, permutation_list)
            for group_index in range(self.groups):
                U[group_index] -= torch.matmul(
                    q_groups[group_index].unsqueeze(1),
                    self.quant_input[group_index, :, permutation_list[group_index][t]].unsqueeze(0),
                )

        del self.float_input
        del self.quant_input



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
SPDX-License-Identifier: MIT
"""

import torch


def unet_input_shape(resolution):
    return (4, resolution // 8, resolution // 8)


def generate_latents(seeds, device, dtype, input_shape):
    """
    Generate a concatenation of latents of a given input_shape
    (batch size excluded) on a target device from one or more seeds.
    """
    latents = None
    if not isinstance(seeds, (list, tuple)):
        seeds = [seeds]
    for seed in seeds:
        generator = torch.Generator(device=device)
        generator = generator.manual_seed(seed)
        image_latents = torch.randn((1, *input_shape),
                                    generator=generator,
                                    device=device,
                                    dtype=dtype)
        latents = image_latents if latents is None else torch.cat((latents, image_latents))
    return latents


def generate_unet_rand_inputs(
        embedding_shape,
        unet_input_shape,
        batch_size=1,
        device='cpu',
        dtype=torch.float32,
        with_return_dict_false=False):
    sample = torch.randn(batch_size, *unet_input_shape, device=device, dtype=dtype)
    unet_rand_inputs = {
        'sample':
            sample,
        'timestep':
            torch.tensor(1, dtype=torch.int64, device=device),
        'encoder_hidden_states':
            torch.randn(batch_size, *embedding_shape, device=device, dtype=dtype)}
    if with_return_dict_false:
        unet_rand_inputs['return_dict'] = False
    return unet_rand_inputs


def generate_unet_21_rand_inputs(
        embedding_shape,
        unet_input_shape,
        batch_size=1,
        device='cpu',
        dtype=torch.float32,
        with_return_dict_false=False):
    unet_rand_inputs = generate_unet_rand_inputs(
        embedding_shape, unet_input_shape, batch_size, device, dtype, with_return_dict_false)
    return tuple(unet_rand_inputs.values())


def generate_unet_xl_rand_inputs(
        embedding_shape,
        unet_input_shape,
        batch_size=1,
        device='cpu',
        dtype=torch.float32,
        with_return_dict_false=False):
    # We need to pass a combination of args and kwargs to ONNX export
    # If we pass all kwargs, something breaks
    # If we pass only the last element as kwargs, since it is a dict, it has a weird interaction and something breaks
    # The solution is to pass only one argument as args, and everything else as kwargs
    unet_rand_inputs = generate_unet_rand_inputs(
        embedding_shape, unet_input_shape, batch_size, device, dtype, with_return_dict_false)
    sample = unet_rand_inputs['sample']
    del unet_rand_inputs['sample']
    unet_rand_inputs['timestep_cond'] = None
    unet_rand_inputs['cross_attention_kwargs'] = None
    unet_rand_inputs['added_cond_kwargs'] = {
        "text_embeds": torch.randn(1, 1280, dtype=dtype, device=device),
        "time_ids": torch.randn(1, 6, dtype=dtype, device=device)}
    inputs = (sample, unet_rand_inputs)
    return inputs



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
from itertools import product
import random
from types import SimpleNamespace

import numpy as np
import pandas as pd
import torch
import torch.backends.cudnn as cudnn
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed

from brevitas import __version__ as brevitas_version
from brevitas import config
from brevitas import torch_version
from brevitas_examples.imagenet_classification.ptq.ptq_common import apply_bias_correction
from brevitas_examples.imagenet_classification.ptq.ptq_common import calibrate
from brevitas_examples.imagenet_classification.ptq.utils import get_model_config
from brevitas_examples.imagenet_classification.ptq.utils import get_quant_model
from brevitas_examples.imagenet_classification.utils import generate_dataloader
from brevitas_examples.imagenet_classification.utils import SEED
from brevitas_examples.imagenet_classification.utils import validate

config.IGNORE_MISSING_KEYS = True

# Manually defined quantized model with original floating point accuracy
IMGCLSMOB_TOP1_MAP = {'quant_mobilenet_v1': 73.390}

parser = argparse.ArgumentParser(description='PyTorch ImageNet PTQ Validation')
parser.add_argument(
    '--calibration-dir',
    default='/scratch/datasets/imagenet_symlink/calibration',
    help='path to folder containing Imagenet calibration folder')
parser.add_argument(
    '--validation-dir',
    default='/scratch/datasets/imagenet_symlink/val',
    help='path to folder containing Imagenet validation folder')

parser.add_argument('--workers', default=8, type=int, help='Number of data loading workers')
parser.add_argument(
    '--batch-size-calibration', default=64, type=int, help='Minibatch size for calibration')
parser.add_argument(
    '--batch-size-validation', default=256, type=int, help='Minibatch size for validation')
parser.add_argument('--gpu', default=None, type=int, help='GPU id to use.')
parser.add_argument('--calibration-samples', default=1000, type=int, help='Calibration size')


def main():
    args = parser.parse_args()
    random.seed(SEED)
    np.random.seed(SEED)
    torch.manual_seed(SEED)

    quant_model_df = pd.DataFrame(
        columns=[
            'Model',
            'Bias correction',
            'Top 1% floating point accuracy',
            'Top 1% quant accuracy',
            'Floating point accuracy - quant accuracy',
            'Quant accuracy / floating point accuracy',
            'Calibration size',
            'Calibration batch size',
            'Torch version',
            'Brevitas version'])

    ptq_quant_models(quant_model_df, args)


def ptq_quant_models(df, args):
    options = [
        IMGCLSMOB_TOP1_MAP.keys(),
        [8],  #Bit width
        [True],  # Bias Correction
    ]

    combinations = list(product(*options))
    k = 0
    for i, (model_name, bit_width, bias_corr) in enumerate(combinations):
        fp_accuracy = IMGCLSMOB_TOP1_MAP[model_name]

        # Get model-specific configurations about input shapes and normalization
        model_config = get_model_config(model_name)

        # Generate calibration and validation dataloaders
        resize_shape = model_config['resize_shape']
        center_crop_shape = model_config['center_crop_shape']
        inception_preprocessing = model_config['inception_preprocessing']
        calib_loader = generate_dataloader(
            args.calibration_dir,
            args.batch_size_calibration,
            args.workers,
            resize_shape,
            center_crop_shape,
            args.calibration_samples,
            inception_preprocessing)
        val_loader = generate_dataloader(
            args.validation_dir,
            args.batch_size_validation,
            args.workers,
            resize_shape,
            center_crop_shape,
            inception_preprocessing=inception_preprocessing)

        # Get the model from torchvision
        model = get_quant_model(model_name, bit_width=bit_width)

        if args.gpu is not None:
            torch.cuda.set_device(args.gpu)
            model = model.cuda(args.gpu)
            cudnn.benchmark = False

        # Calibrate the model on the calibration dataloader
        calibrate(calib_loader, model)

        if bias_corr:
            apply_bias_correction(calib_loader, model)

        # Validate the model on the validation dataloader
        top1 = validate(val_loader, model)

        # Generate metrics for benchmarking
        top1 = np.around(top1, decimals=3)
        acc_diff = np.around(top1 - fp_accuracy, decimals=3)
        acc_ratio = np.around(top1 / fp_accuracy, decimals=3)

        df.loc[k] = [
            model_name,
            bias_corr,
            fp_accuracy,
            top1,
            acc_diff,
            acc_ratio,
            args.calibration_samples,
            args.batch_size_calibration,
            torch_version,
            brevitas_version]
        k += 1
    df.to_csv('RESULTS_IMGCLSMOB.csv', index=False, mode='w')


if __name__ == '__main__':
    main()



import brevitas.nn as qnn
from brevitas_examples.common.generative.quantizers import IntWeightSymmetricGroupQuant
from brevitas_examples.llm.llm_quant.export import export_packed_onnx
import torch.nn as nn
import torch


class Model(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = qnn.QuantLinear(32, 32, weight_quant=IntWeightSymmetricGroupQuant, weight_bit_width=4, weight_group_size=8)

    def forward(self, x):
        return self.linear(x)

model = Model()
model(torch.randn(1, 32))
model.eval()
export_packed_onnx(model, torch.randn(1,32), 'test.onnx')


# from functools import partial
# shape_dict = dict()
# def hook(module, input, output, name):
#     shape_dict[id(module)] = output.shape

# hook_list = []
# for name, module in pipe.unet.named_modules():
#     if isinstance(module, (torch.nn.Linear, torch.nn.Conv2d)):
#         partial_hook = partial(hook, name=name)
#         hook_list.append(module.register_forward_hook(partial_hook))

# ### FORWARD CALL HERE
# pipe(prompt, ...)
# ### END FORWARD CALL

# for name, shape in shape_dict.items():
#     print(name, shape)


"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""
import re

import torch
from torch import nn

from brevitas import nn as qnn
from brevitas.core.zero_point import ParameterFromStatsFromParameterZeroPoint
from brevitas.graph.quantize import layerwise_quantize
from brevitas.quant.experimental.float import Fp8e4m3Act
from brevitas.quant.experimental.float import Fp8e4m3ActPerTensorFloat
from brevitas.quant.experimental.float import Fp8e4m3WeightPerChannelFloat
from brevitas.quant.experimental.float import Fp8e4m3WeightPerTensorFloat
from brevitas.quant.experimental.float_quant_fnuz import Fp8e4m3FNUZActPerTensorFloat
from brevitas.quant.experimental.float_quant_fnuz import Fp8e4m3FNUZWeightPerChannelFloat
from brevitas.quant.experimental.float_quant_fnuz import Fp8e4m3FNUZWeightPerTensorFloat
from brevitas.quant.experimental.float_quant_ocp import Fp8e4m3OCPActPerTensorFloat
from brevitas.quant.experimental.float_quant_ocp import Fp8e4m3OCPWeightPerChannelFloat
from brevitas.quant.experimental.float_quant_ocp import Fp8e4m3OCPWeightPerTensorFloat
from brevitas.quant.experimental.mx_quant_ocp import MXFloat8e4m3Act
from brevitas.quant.experimental.mx_quant_ocp import MXFloat8e4m3Weight
from brevitas.quant.experimental.mx_quant_ocp import MXFloat8e4m3WeightMSE
from brevitas.quant.experimental.mx_quant_ocp import MXInt8Act
from brevitas.quant.experimental.mx_quant_ocp import MXInt8Weight
from brevitas.quant.experimental.mx_quant_ocp import MXInt8WeightMSE
from brevitas.quant.experimental.mx_quant_ocp import ShiftedMXUInt8Weight
from brevitas.quant.experimental.mx_quant_ocp import ShiftedMXUInt8WeightMSE
from brevitas.quant.fixed_point import Int8ActPerTensorFixedPoint
from brevitas.quant.fixed_point import Int8ActPerTensorFixedPointMSE
from brevitas.quant.fixed_point import Int8WeightPerChannelFixedPoint
from brevitas.quant.fixed_point import Int8WeightPerChannelFixedPointMSE
from brevitas.quant.fixed_point import Int8WeightPerTensorFixedPoint
from brevitas.quant.fixed_point import Int8WeightPerTensorFixedPointMSE
from brevitas.quant.scaled_int import Int8ActPerTensorFloat
from brevitas.quant.scaled_int import Int8ActPerTensorFloatMSE
from brevitas.quant.scaled_int import Int8WeightPerChannelFloat
from brevitas.quant.scaled_int import Int8WeightPerChannelFloatHQO
from brevitas.quant.scaled_int import Int8WeightPerChannelFloatMSE
from brevitas.quant.scaled_int import Int8WeightPerTensorFloat
from brevitas.quant.scaled_int import Int8WeightPerTensorFloatHQO
from brevitas.quant.scaled_int import Int8WeightPerTensorFloatMSE
from brevitas.quant.shifted_scaled_int import ShiftedUint8ActPerTensorFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8ActPerTensorFloatMSE
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightGroupQuantFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerChannelFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerChannelFloatHQO
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerChannelFloatMSE
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerGroupFloatHQO
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerTensorFloat
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerTensorFloatHQO
from brevitas.quant.shifted_scaled_int import ShiftedUint8WeightPerTensorFloatMSE
from brevitas_examples.common.generative.nn import LoRACompatibleQuantConv2d
from brevitas_examples.common.generative.nn import LoRACompatibleQuantLinear
from brevitas_examples.common.generative.quantizers import Fp8e4m3DynamicActPerGroupFloat
from brevitas_examples.common.generative.quantizers import Fp8e4m3WeightSymmetricGroupQuant
from brevitas_examples.common.generative.quantizers import Int8DynamicActPerGroupFloat
from brevitas_examples.common.generative.quantizers import Int8DynamicActPerRowFloat
from brevitas_examples.common.generative.quantizers import Int8DynamicActPerTensorFloat
from brevitas_examples.common.generative.quantizers import IntWeightSymmetricGroupQuant
from brevitas_examples.common.generative.quantizers import ShiftedUint8DynamicActPerRowFloat
from brevitas_examples.common.generative.quantizers import ShiftedUint8DynamicActPerTensorFloat

WEIGHT_QUANT_MAP = {
    'int': {
        'float_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFloat, 'asym': ShiftedUint8WeightPerTensorFloat},
                'per_channel': {
                    'sym': Int8WeightPerChannelFloat, 'asym': ShiftedUint8WeightPerChannelFloat},
                'per_group': {
                    'sym': IntWeightSymmetricGroupQuant,
                    'asym': ShiftedUint8WeightGroupQuantFloat}},
            'mse': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFloatMSE,
                    'asym': ShiftedUint8WeightPerTensorFloatMSE},
                'per_channel': {
                    'sym': Int8WeightPerChannelFloatMSE,
                    'asym': ShiftedUint8WeightPerChannelFloatMSE}},
            'hqo': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFloatHQO,
                    'asym': ShiftedUint8WeightPerTensorFloatHQO},
                'per_channel': {
                    'sym': Int8WeightPerChannelFloatHQO,
                    'asym': ShiftedUint8WeightPerChannelFloatHQO},
                'per_group': {
                    'asym': ShiftedUint8WeightPerGroupFloatHQO}},},
        'po2_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFixedPoint},
                'per_channel': {
                    'sym': Int8WeightPerChannelFixedPoint},
                'per_group': {
                    'sym': MXInt8Weight, 'asym': ShiftedMXUInt8Weight}},
            'mse': {
                'per_tensor': {
                    'sym': Int8WeightPerTensorFixedPointMSE},
                'per_channel': {
                    'sym': Int8WeightPerChannelFixedPointMSE},
                'per_group': {
                    'sym': MXInt8WeightMSE, 'asym': ShiftedMXUInt8WeightMSE}}}},
    'float': {
        'float_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Fp8e4m3WeightPerTensorFloat},
                'per_channel': {
                    'sym': Fp8e4m3WeightPerChannelFloat},
                'per_group': {
                    'sym': Fp8e4m3WeightSymmetricGroupQuant}}}},
    'float_ocp': {
        'float_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Fp8e4m3OCPWeightPerTensorFloat},
                'per_channel': {
                    'sym': Fp8e4m3OCPWeightPerChannelFloat}}},
        'po2_scale': {
            'stats': {
                'per_group': {
                    'sym': MXFloat8e4m3Weight}},
            'mse': {
                'per_group': {
                    'sym': MXFloat8e4m3WeightMSE}}}},
    'float_fnuz': {
        'float_scale': {
            'stats': {
                'per_tensor': {
                    'sym': Fp8e4m3FNUZWeightPerTensorFloat},
                'per_channel': {
                    'sym': Fp8e4m3FNUZWeightPerChannelFloat}}}}}

INPUT_QUANT_MAP = {
    'int': {
        'static': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFloat, 'asym': ShiftedUint8ActPerTensorFloat}},
                'mse': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFloatMSE,
                        'asym': ShiftedUint8ActPerTensorFloatMSE}}},
            'po2_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFixedPoint}},
                'mse': {
                    'per_tensor': {
                        'sym': Int8ActPerTensorFixedPointMSE}}}},
        'dynamic': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Int8DynamicActPerTensorFloat,
                        'asym': ShiftedUint8DynamicActPerTensorFloat},
                    'per_row': {
                        'sym': Int8DynamicActPerRowFloat,
                        'asym': ShiftedUint8DynamicActPerRowFloat},
                    'per_group': {
                        'sym': Int8DynamicActPerGroupFloat}}},
            'po2_scale': {
                'stats': {
                    'per_group': {
                        'sym': MXInt8Act}}}}},
    'float': {
        'static': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Fp8e4m3ActPerTensorFloat}}}},
        'dynamic': {
            'float_scale': {
                'stats': {
                    'per_group': Fp8e4m3DynamicActPerGroupFloat}}},
        'no_scale': {
            'sym': Fp8e4m3Act,}},
    'float_ocp': {
        'static': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Fp8e4m3OCPActPerTensorFloat}}}},
        'dynamic': {
            'po2_scale': {
                'stats': {
                    'per_group': {
                        'sym': MXFloat8e4m3Act}}}}},
    'float_fnuz': {
        'static': {
            'float_scale': {
                'stats': {
                    'per_tensor': {
                        'sym': Fp8e4m3FNUZActPerTensorFloat}}}}}}


def generate_quantizers(
        dtype,
        weight_bit_width,
        weight_param_method,
        weight_scale_precision,
        weight_quant_type,
        weight_quant_granularity,
        weight_group_size,
        quantize_weight_zero_point,
        weight_quant_format='int',
        weight_group_dim=None,
        input_bit_width=None,
        input_quant_format='',
        input_scale_precision=None,
        input_scale_type=None,
        input_param_method=None,
        input_quant_type=None,
        input_quant_granularity=None,
        input_group_size=None,
        quantize_input_zero_point=False,
        device=None,
        weight_kwargs=None,
        input_kwargs=None):
    """
    Replace float layers with quant layers in the target model
    """
    # Retrive base input and weight quantizers
    # match against custom float format
    if re.compile(r'e[1-8]m[1-8]').findall(weight_quant_format):
        format = re.compile(r'e[1-8]m[1-8]').findall(weight_quant_format)[0]
        weight_quant_format = weight_quant_format.replace('_' + format, '')
        weight_float_format = {
            'exponent_bit_width': int(format[1]), 'mantissa_bit_width': int(format[3])}
    else:
        weight_float_format = {}
    if re.compile(r'e[1-8]m[1-8]').findall(input_quant_format):
        format = re.compile(r'e[1-8]m[1-8]').findall(input_quant_format)[0]
        input_quant_format = input_quant_format.replace('_' + format, '')
        input_float_format = {
            'exponent_bit_width': int(format[1]), 'mantissa_bit_width': int(format[3])}
    else:
        input_float_format = {}

    weight_quant = WEIGHT_QUANT_MAP[weight_quant_format][weight_scale_precision][
        weight_param_method][weight_quant_granularity][weight_quant_type]

    if input_bit_width is not None and input_scale_type == 'no_scale':
        input_quant = sym_input_quant = linear_input_quant = INPUT_QUANT_MAP[input_quant_format][
            input_scale_type][input_quant_type]
    elif input_bit_width is not None:
        input_quant = INPUT_QUANT_MAP[input_quant_format][input_scale_type][input_scale_precision][
            input_param_method][input_quant_granularity][input_quant_type]
        # Some activations in MHA should always be symmetric
        sym_input_quant = INPUT_QUANT_MAP[input_quant_format][input_scale_type][
            input_scale_precision][input_param_method][input_quant_granularity]['sym']
        linear_input_quant = INPUT_QUANT_MAP[input_quant_format][input_scale_type][
            input_scale_precision][input_param_method][input_quant_granularity][input_quant_type]

        if input_kwargs is None:
            input_kwargs = dict()

        input_quant = input_quant.let(**input_kwargs)
        sym_input_quant = sym_input_quant.let(**input_kwargs)
        linear_input_quant = linear_input_quant.let(**input_kwargs)

    else:
        input_quant = None
        sym_input_quant = None
        linear_input_quant = None

    # Modify the weight quantizer based on the arguments passed in
    weight_quant = weight_quant.let(
        **{
            'bit_width': weight_bit_width,
            'narrow_range': False,
            'quantize_zero_point': quantize_weight_zero_point},
        **weight_float_format)

    if weight_group_dim is not None:
        weight_quant = weight_quant.let(**{'group_dim': weight_group_dim})

    if dtype == torch.float16:
        weight_quant = weight_quant.let(**{'scaling_min_val': 1e-4})
    if weight_kwargs is not None:
        weight_quant = weight_quant.let(**weight_kwargs)

    # Set the group_size is we're doing groupwise quantization
    if weight_quant_granularity == 'per_group':
        weight_quant = weight_quant.let(**{'group_size': weight_group_size})
    # weight scale is converted to a standalone parameter

    weight_quant = weight_quant.let(scaling_impl_type='parameter_from_stats')
    # weight zero-point is converted to a standalone parameter
    # This is done already by default in the per_group quantizer
    if weight_quant_type == 'asym' and weight_quant_granularity != 'per_group':
        weight_quant = weight_quant.let(zero_point_impl=ParameterFromStatsFromParameterZeroPoint)

    # Modify the input quantizers based on the arguments passed in
    if input_quant is not None:
        input_quant = input_quant.let(
            **{
                'bit_width': input_bit_width,
                'quantize_zero_point': quantize_input_zero_point,
                'dtype': dtype,
                'device': device},
            **input_float_format)
        if input_scale_type == 'dynamic':
            if input_quant_granularity == 'per_row':
                input_quant = input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_fn': lambda x,
                                                            shape: x.view(*shape[:-1], 1),
                        'stats_reduce_dim': 1})
            elif input_quant_granularity == 'per_group':
                input_quant = input_quant.let(**{'group_dim': 2, 'group_size': input_group_size})
    if sym_input_quant is not None:
        sym_input_quant = sym_input_quant.let(
            **{
                'bit_width': input_bit_width,
                'quantize_zero_point': quantize_input_zero_point,
                'dtype': dtype,
                'device': device},
            **input_float_format)
        if input_scale_type == 'dynamic':
            if input_quant_granularity == 'per_tensor':
                q_scaled_quant = sym_input_quant
                k_transposed_quant = sym_input_quant
            elif input_quant_granularity == 'per_row':
                q_scaled_quant = sym_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_fn': lambda x,
                                                            shape: x.view(*shape[:-1], 1),
                        'permute_dims': None,
                        'stats_reduce_dim': 1})
                k_transposed_quant = sym_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_fn':
                            lambda x,
                            shape: x.view(shape[0], 1, shape[-1]),
                        'permute_dims': (0, 2, 1),
                        'stats_reduce_dim':
                            1})
            elif input_quant_granularity == 'per_group':
                q_scaled_quant = sym_input_quant.let(
                    **{
                        'group_dim': 2, 'group_size': input_group_size})
                k_transposed_quant = sym_input_quant.let(
                    **{
                        'group_dim': 1, 'group_size': input_group_size})
            v_quant = q_scaled_quant
            attn_output_weights_quant = q_scaled_quant
        else:
            q_scaled_quant = v_quant = k_transposed_quant = attn_output_weights_quant = sym_input_quant
    else:
        q_scaled_quant = v_quant = k_transposed_quant = attn_output_weights_quant = None
    if linear_input_quant is not None:
        linear_input_quant = linear_input_quant.let(
            **{
                'bit_width': input_bit_width,
                'quantize_zero_point': quantize_input_zero_point,
                'dtype': dtype,
                'device': device},
            **input_float_format)
        if input_scale_type == 'dynamic':
            if input_quant_granularity == 'per_row':
                linear_input_quant = linear_input_quant.let(
                    **{
                        'dynamic_scaling_broadcastable_fn': lambda x,
                                                            shape: x.view(*shape[:-1], 1),
                        'permute_dims': None,
                        'stats_reduce_dim': 1})
            elif input_quant_granularity == 'per_group':
                linear_input_quant = linear_input_quant.let(
                    **{
                        'group_dim': -1, 'group_size': input_group_size})
    return linear_input_quant, weight_quant, input_quant, q_scaled_quant, k_transposed_quant, v_quant, attn_output_weights_quant


def generate_quant_maps(
        linear_input_quant,
        weight_quant,
        input_quant,
        q_scaled_quant,
        k_transposed_quant,
        v_quant,
        attn_output_weights_quant,
        dtype,
        device,
        input_quant_format,
        quantize_embedding):

    quant_linear_kwargs = {
        'input_quant': linear_input_quant,
        'weight_quant': weight_quant,
        'dtype': dtype,
        'device': device}
    quant_conv_kwargs = {
        'input_quant': input_quant, 'weight_quant': weight_quant, 'dtype': dtype, 'device': device}

    quant_mha_kwargs = {
        'in_proj_input_quant': input_quant,
        'in_proj_weight_quant': weight_quant,
        'in_proj_bias_quant': None,
        'softmax_input_quant': None,
        'attn_output_weights_quant': attn_output_weights_quant,
        'attn_output_weights_signed': 'float' in input_quant_format,
        'q_scaled_quant': q_scaled_quant,
        'k_transposed_quant': k_transposed_quant,
        'v_quant': v_quant,
        'out_proj_input_quant': input_quant,
        'out_proj_weight_quant': weight_quant,
        'out_proj_bias_quant': None,
        'out_proj_output_quant': None,
        'batch_first': True,
        # activation equalization requires packed_in_proj
        # since it supports only self-attention
        'packed_in_proj': True,
        'dtype': dtype,
        'device': device}

    layer_map = {
        nn.Linear: (qnn.QuantLinear, quant_linear_kwargs),
        nn.Conv2d: (qnn.QuantConv2d, quant_conv_kwargs),
        'diffusers.models.lora.LoRACompatibleLinear':
            (LoRACompatibleQuantLinear, quant_linear_kwargs),
        'diffusers.models.lora.LoRACompatibleConv': (LoRACompatibleQuantConv2d, quant_conv_kwargs),
        nn.MultiheadAttention: (qnn.QuantMultiheadAttention, quant_mha_kwargs)}

    if quantize_embedding:
        quant_embedding_kwargs = {'weight_quant': weight_quant, 'dtype': dtype, 'device': device}
        layer_map[nn.Embedding] = (qnn.QuantEmbedding, quant_embedding_kwargs)
    return layer_map


def quantize_model(
        model,
        dtype,
        weight_bit_width,
        weight_param_method,
        weight_scale_precision,
        weight_quant_type,
        weight_quant_granularity,
        weight_group_size,
        quantize_weight_zero_point,
        weight_quant_format='int',
        name_blacklist=None,
        input_bit_width=None,
        input_quant_format='',
        input_scale_precision=None,
        input_scale_type=None,
        input_param_method=None,
        input_quant_type=None,
        input_quant_granularity=None,
        input_group_size=None,
        quantize_input_zero_point=False,
        quantize_embedding=False,
        device=None,
        weight_kwargs=None,
        input_kwargs=None):

    linear_input_quant, weight_quant, input_quant, q_scaled_quant, k_transposed_quant, v_quant, attn_output_weights_quant = generate_quantizers(
        dtype,
        weight_bit_width,
        weight_param_method,
        weight_scale_precision,
        weight_quant_type,
        weight_quant_granularity,
        weight_group_size,
        quantize_weight_zero_point,
        weight_quant_format,
        input_bit_width,
        input_quant_format,
        input_scale_precision,
        input_scale_type,
        input_param_method,
        input_quant_type,
        input_quant_granularity,
        input_group_size,
        quantize_input_zero_point,
        device,
        weight_kwargs,
        input_kwargs)
    layer_map = generate_quant_maps(
        linear_input_quant,
        weight_quant,
        input_quant,
        q_scaled_quant,
        k_transposed_quant,
        v_quant,
        attn_output_weights_quant,
        dtype,
        device,
        input_quant_format,
        quantize_embedding)
    model = layerwise_quantize(
        model=model, compute_layer_map=layer_map, name_blacklist=name_blacklist)
    return model



from brevitas.inject.enum import ScalingImplType
from brevitas.inject.enum import ScalingPerOutputType

assert ScalingImplType
assert ScalingPerOutputType

from brevitas.core.stats import SCALAR_SHAPE

from .float_scaling import FloatScaling
from .int_scaling import IntScaling
from .int_scaling import PowerOfTwoIntScaling
from .pre_scaling import AccumulatorAwareParameterPreScaling
from .pre_scaling import AccumulatorAwareZeroCenterParameterPreScaling
from .pre_scaling import ParameterPreScalingWeightNorm
from .runtime import RuntimeStatsScaling
from .runtime import StatsFromParameterScaling
from .standalone import ConstScaling
from .standalone import ParameterFromRuntimeStatsScaling
from .standalone import ParameterFromStatsFromParameterScaling
from .standalone import ParameterScaling

SCALING_STATS_REDUCE_DIM = 1



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .binary import *
from .fixed_point import *
from .none import *
from .scaled_int import *
from .shifted_scaled_int import *
from .ternary import *



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from torch import Tensor
import torch.nn as nn

import brevitas.nn as qnn
from brevitas.nn.quant_layer import WeightQuantType

from .common import CommonIntAccumulatorAwareWeightQuant
from .common import CommonIntWeightPerChannelQuant
from .common import CommonUintActQuant
from .common import ConstUint8ActQuant

__all__ = [
    "float_espcn", "quant_espcn", "quant_espcn_a2q", "quant_espcn_base", "FloatESPCN", "QuantESPCN"]

IO_DATA_BIT_WIDTH = 8
IO_ACC_BIT_WIDTH = 32


def weight_init(layer):
    if isinstance(layer, nn.Conv2d):
        nn.init.kaiming_normal_(layer.weight, nn.init.calculate_gain('relu'))
        if layer.bias is not None:
            layer.bias.data.zero_()


class FloatESPCN(nn.Module):
    """Floating-point version of Efficient Sub-Pixel Convolution Network (ESPCN)"""

    def __init__(self, upscale_factor: int = 3, num_channels: int = 3):
        super(FloatESPCN, self).__init__()
        self.upscale_factor = upscale_factor

        self.conv1 = nn.Conv2d(
            in_channels=num_channels,
            out_channels=64,
            kernel_size=5,
            stride=1,
            padding=2,
            bias=True)
        self.conv2 = nn.Conv2d(
            in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=True)
        self.conv3 = nn.Conv2d(
            in_channels=64, out_channels=32, kernel_size=3, stride=1, padding=1, bias=True)
        self.conv4 = nn.Conv2d(
            in_channels=32,
            out_channels=num_channels * pow(upscale_factor, 2),
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False)
        self.pixel_shuffle = nn.PixelShuffle(upscale_factor)

        self.bn1 = nn.BatchNorm2d(64)
        self.bn2 = nn.BatchNorm2d(64)
        self.bn3 = nn.BatchNorm2d(32)

        self.relu = nn.ReLU(inplace=True)
        self.out = nn.ReLU(inplace=True)  # To mirror quant version

        # Initialize weights
        self.apply(weight_init)

    def forward(self, inp: Tensor):
        x = torch.relu(inp)  # Adding for finn-onnx compatability
        x = self.relu(self.bn1(self.conv1(x)))
        x = self.relu(self.bn2(self.conv2(x)))
        x = self.relu(self.bn3(self.conv3(x)))
        x = self.pixel_shuffle(self.conv4(x))
        x = self.out(x)  # To mirror quant version
        return x


class QuantESPCN(FloatESPCN):
    """FINN-Friendly Quantized Efficient Sub-Pixel Convolution Network (ESPCN)"""

    def __init__(
            self,
            upscale_factor: int = 3,
            num_channels: int = 3,
            weight_bit_width: int = 4,
            act_bit_width: int = 4,
            acc_bit_width: int = 32,
            weight_quant: WeightQuantType = CommonIntWeightPerChannelQuant):
        super(QuantESPCN, self).__init__(upscale_factor=upscale_factor)

        # Quantizing the activations to all conv2d layers to unsigned because they
        # are all preceded by ReLUs, which have non-negative ranges

        # Quantizing input quant conv2d to use 8-bit inputs and weights with
        # an accumulator size of 32 bits
        self.conv1 = qnn.QuantConv2d(
            in_channels=num_channels,
            out_channels=64,
            kernel_size=5,
            stride=1,
            padding=2,
            bias=True,
            input_bit_width=IO_DATA_BIT_WIDTH,
            input_quant=CommonUintActQuant,
            weight_bit_width=IO_DATA_BIT_WIDTH,
            weight_accumulator_bit_width=IO_ACC_BIT_WIDTH,
            weight_quant=weight_quant)
        self.conv2 = qnn.QuantConv2d(
            in_channels=64,
            out_channels=64,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
            input_bit_width=act_bit_width,
            input_quant=CommonUintActQuant,
            weight_bit_width=weight_bit_width,
            weight_accumulator_bit_width=acc_bit_width,
            weight_quant=weight_quant)
        self.conv3 = qnn.QuantConv2d(
            in_channels=64,
            out_channels=32,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
            input_bit_width=act_bit_width,
            input_quant=CommonUintActQuant,
            weight_bit_width=weight_bit_width,
            weight_accumulator_bit_width=acc_bit_width,
            weight_quant=weight_quant)
        # We quantize the weights and input activations of the final layer
        # to 8-bit integers. We do not apply the accumulator constraint to
        # the final convolution layer. FINN does not currently support
        # per-tensor quantization or biases for sub-pixel convolution layers.
        self.conv4 = qnn.QuantConv2d(
            in_channels=32,
            out_channels=num_channels * pow(upscale_factor, 2),
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False,
            input_bit_width=act_bit_width,
            input_quant=CommonUintActQuant,
            weight_bit_width=IO_DATA_BIT_WIDTH,
            weight_quant=CommonIntWeightPerChannelQuant,
            weight_scaling_per_output_channel=False)

        self.bn1 = nn.BatchNorm2d(64)
        self.bn2 = nn.BatchNorm2d(64)
        self.bn3 = nn.BatchNorm2d(32)

        self.relu = nn.ReLU(inplace=True)
        # Using a QuantReLU here because we need to read out a uint8 image, but FINN
        # requires a ReLU node to precede an unsigned int quant node
        self.out = qnn.QuantReLU(act_quant=ConstUint8ActQuant, bit_width=IO_DATA_BIT_WIDTH)

        # Initialize weights
        self.apply(weight_init)


def float_espcn(upscale_factor: int, num_channels: int = 3) -> FloatESPCN:
    """ """
    return FloatESPCN(upscale_factor, num_channels=num_channels)


def quant_espcn(
        upscale_factor: int,
        num_channels: int = 3,
        weight_bit_width: int = 8,
        act_bit_width: int = 8,
        acc_bit_width: int = 32,
        weight_quant: WeightQuantType = CommonIntWeightPerChannelQuant) -> QuantESPCN:
    """ """
    return QuantESPCN(
        upscale_factor=upscale_factor,
        num_channels=num_channels,
        act_bit_width=act_bit_width,
        acc_bit_width=acc_bit_width,
        weight_bit_width=weight_bit_width,
        weight_quant=weight_quant)


def quant_espcn_a2q(
        upscale_factor: int, weight_bit_width: int, act_bit_width: int, acc_bit_width: int):
    """Integer-quantized FINN-friendly ESPCN model for BSD300 using
    the accumulator-aware weight quantizer"""
    return QuantESPCN(
        upscale_factor=upscale_factor,
        act_bit_width=act_bit_width,
        acc_bit_width=acc_bit_width,
        weight_bit_width=weight_bit_width,
        weight_quant=CommonIntAccumulatorAwareWeightQuant)


def quant_espcn_base(upscale_factor: int, weight_bit_width: int, act_bit_width: int):
    """Integer-quantized FINN-friendly ESPCN model for BSD300 using
    a vanilla per-channel weight quantizer"""
    return QuantESPCN(
        upscale_factor=upscale_factor,
        act_bit_width=act_bit_width,
        weight_bit_width=weight_bit_width,
        weight_quant=CommonIntWeightPerChannelQuant)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from torch.autograd import Function

from brevitas.nn.mixin.base import QuantLayerMixin
from brevitas.proxy.quant_proxy import QuantProxyProtocol
from brevitas.quant_tensor import QuantTensor


class DebugMarkerFunction(Function):

    @staticmethod
    def symbolic(g, input, export_debug_name):
        ret = g.op('brevitas.onnx::DebugMarker', input, export_debug_name_s=export_debug_name)
        ret.setType(input.type())
        return ret

    @staticmethod
    def forward(ctx, input, export_debug_name):
        return input


class ONNXDebugHook(object):

    def __init__(self, input_enabled, output_enabled):
        self.values = {}
        self.input_enabled = input_enabled
        self.output_enabled = output_enabled

    def unpack(self, value):
        if isinstance(value, tuple) and len(value) == 1:
            return value[0]
        else:
            return value

    def __call__(self, module, module_in, module_out):
        if self.input_enabled:
            self.values[module.export_debug_name + ".input"] = self.unpack(module_in)
        if self.output_enabled:
            self.values[module.export_debug_name + ".output"] = self.unpack(module_out)

    def clear(self):
        self.values = {}


def enable_debug(model, input_enabled=True, output_enabled=True, proxy_level=False):
    base_filter_class = QuantProxyProtocol if proxy_level else QuantLayerMixin
    filter_fn = lambda x: isinstance(x, base_filter_class)
    hook = ONNXDebugHook(input_enabled, output_enabled)
    for name, module in model.named_modules():
        if hasattr(module, "export_debug_name") and filter_fn(module):
            module.export_debug_name = name
            module.export_input_debug = input_enabled
            module.export_output_debug = output_enabled
            module.register_forward_hook(hook)
    return hook



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, Type, Union

from torch.nn import Module

from brevitas.inject import BaseInjector as Injector
from brevitas.proxy.runtime_quant import AccQuantProxyProtocol
from brevitas.quant_tensor import QuantTensor

from .mixin.acc import AccQuantType
from .mixin.acc import QuantClampMixin
from .mixin.acc import TruncMixin
from .mixin.base import QuantLayerMixin


class TruncQuantAccumulator(TruncMixin, QuantLayerMixin, Module):

    def __init__(
            self,
            trunc_quant: Optional[AccQuantType] = None,
            return_quant_tensor: bool = True,
            **kwargs):
        QuantLayerMixin.__init__(self, return_quant_tensor)
        TruncMixin.__init__(self, trunc_quant=trunc_quant, **kwargs)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return True

    def forward(self, input: QuantTensor):
        x = self.unpack_input(input)
        x = self.trunc_quant(x)
        return self.pack_output(x)


class ClampQuantAccumulator(QuantClampMixin, QuantLayerMixin, Module):

    def __init__(
            self,
            clamp_quant: Optional[AccQuantType] = None,
            return_quant_tensor: bool = True,
            **kwargs):
        QuantLayerMixin.__init__(self, return_quant_tensor)
        QuantClampMixin.__init__(self, clamp_quant=clamp_quant, **kwargs)

    @property
    def channelwise_separable(self) -> bool:
        return True

    @property
    def requires_export_handler(self):
        return True

    def forward(self, input: QuantTensor):
        x = self.unpack_input(input)
        x = self.clamp_quant(x)
        return self.pack_output(x)



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod
import inspect
from inspect import getcallargs

import torch
from torch.nn import Module
from torch.overrides import get_testing_overrides

from brevitas.fx import GraphModule
from brevitas.fx import immutable_dict
from brevitas.fx import Node
from brevitas.graph.utils import *
from brevitas.utils.python_utils import islambda

__all__ = [
    'Transform',
    'PerInputTransform',
    'GraphTransform',
    'PerInputModuleToModuleByHook',
    'ModuleToModule',
    'InsertModuleCallAfter',
    'ModuleToModuleByName',
    'ModuleToModuleByClass',
    'ModuleInstanceToModuleInstance',
    'ModuleToModuleByInstance',
    'MethodToModule',
    'FnToModule',
    'CallableToModule']

_TORCH_TESTING_DICT = get_testing_overrides()


class Transform(ABC):

    @abstractmethod
    def apply(self, model: Module) -> Module:
        pass


class PerInputTransform(ABC):

    @abstractmethod
    def apply(self, model: Module, inp: torch.Tensor) -> Module:
        pass


class GraphTransform(Transform):

    @abstractmethod
    def apply(self, graph_model: GraphModule) -> GraphModule:
        pass


class UntilFixedPointGraphTransform(Transform):

    @abstractmethod
    def is_converged(self, graph_model: GraphModule) -> bool:
        pass

    def apply(self, graph_model: GraphModule) -> GraphModule:
        while not self.is_converged(graph_model):
            continue
        return graph_model


class PerInputModuleToModuleByHook(PerInputTransform, ABC):

    def __init__(self):
        self.input_size_map = {}
        self.hook_handlers = []

    @abstractmethod
    def register_hooks(self, model):
        pass

    @abstractmethod
    def replace_modules(self, model):
        pass

    def hook_fn(self, module, inp):
        if isinstance(inp, tuple):
            assert len(inp) == 1
            inp = inp[0]
        size = inp.size()
        if module in self.input_size_map.keys() and self.input_size_map[module] != size:
            raise RuntimeError("Layer called multiple times with different input sizes.")
        self.input_size_map[module] = size

    def cleanup(self):
        for hook in self.hook_handlers:
            hook.remove()
        self.hook_handlers = []
        self.input_size_map = {}

    def apply(self, model: Module, *model_args, **model_kwargs):
        self.register_hooks(model)
        model(*model_args, **model_kwargs)
        self.replace_modules(model)
        self.cleanup()
        return model


class ModuleToModule(GraphTransform, ABC):

    def __init__(self, new_module_class, **kwargs):
        super().__init__()
        self.new_module_class = new_module_class
        self.new_module_kwargs = kwargs

    def _map_origin_vars(self, vars: dict):
        return {k: v is not None if k == 'bias' else v for k, v in vars.items()}

    def _module_attributes(self, module):
        attrs = vars(module)
        # workaround since bias doesn't show up on vars of Linear
        if hasattr(module, 'bias'):
            attrs['bias'] = module.bias
        return attrs

    def _evaluate_new_kwargs(self, new_kwargs, old_module, name):
        update_dict = dict()
        for k, v in self.new_module_kwargs.items():
            if islambda(v):
                if name is not None:
                    # Two types of lambdas are admitted now, with/without the name of the module as input
                    if len(inspect.getfullargspec(v).args) == 2:
                        v = v(old_module, name)
                    elif len(inspect.getfullargspec(v).args) == 1:
                        v = v(old_module)
                else:
                    v = v(old_module)
            update_dict[k] = v
        new_kwargs.update(update_dict)
        return new_kwargs

    def _init_new_module(self, old_module: Module, name=None):
        # get attributes of original module
        new_kwargs = self._module_attributes(old_module)
        # transforms attribute of original module, e.g. bias Parameter -> bool
        new_kwargs = self._map_origin_vars(new_kwargs)
        # restrict to only values that are in the init of the new module
        new_module_signature_keys = signature_keys(self.new_module_class)
        new_kwargs = {k: v for k, v in new_kwargs.items() if k in new_module_signature_keys}
        # update with kwargs passed to the rewriter
        new_kwargs = self._evaluate_new_kwargs(new_kwargs, old_module, name)
        # init the new module
        new_module = self.new_module_class(**new_kwargs)
        return new_module

    def _replace_old_module(self, model, old_module, new_module, load_state_dict=True):
        replace_module(model, old_module, new_module)
        if load_state_dict:
            new_module.load_state_dict(old_module.state_dict())


class InsertModuleCallAfter(GraphTransform):

    def __init__(self, module_name, node, node_to_exclude=()):
        self.module_name = module_name
        self.node = node
        self.node_to_exclude = node_to_exclude

    def apply(self, graph_model: GraphModule) -> GraphModule:
        with graph_model.graph.inserting_after(self.node):
            quant_identity_node = graph_model.graph.call_module(self.module_name, args=(self.node,))
        replace_all_uses_except(
            self.node, quant_identity_node, [quant_identity_node] + list(self.node_to_exclude))
        graph_model.recompile()
        graph_model.graph.lint()
        return graph_model


class ModuleInstanceToModuleInstance(Transform):

    def __init__(self, old_module_instance, new_module_instance):
        self.old_module_instance = old_module_instance
        self.new_module_instance = new_module_instance

    def apply(self, model: GraphModule) -> GraphModule:
        for old_module in model.modules():
            if old_module is self.old_module_instance:
                # init the new module based on the old one
                replace_module(model, old_module, self.new_module_instance)
                break
        return model


class ModuleToModuleByName(ModuleToModule):

    def __init__(self, old_module_name, new_module_class, **kwargs):
        super().__init__(new_module_class, **kwargs)
        self.old_module_name = old_module_name

    def apply(self, model: GraphModule) -> GraphModule:
        for name, old_module in model.named_modules():
            if name == self.old_module_name:
                # init the new module based on the old one
                new_module = self._init_new_module(old_module)
                self._replace_old_module(model, old_module, new_module)
                break
        return model


class ModuleToModuleByInstance(ModuleToModule):

    def __init__(self, old_module_instance, new_module_class, **kwargs):
        super().__init__(new_module_class, **kwargs)
        self.old_module_instance = old_module_instance

    def apply(self, model: GraphModule) -> GraphModule:
        for name, old_module in model.named_modules():
            if old_module is self.old_module_instance:
                # init the new module based on the old one
                new_module = self._init_new_module(old_module, name)
                self._replace_old_module(model, old_module, new_module)
                break
        return model


class ModuleToModuleByClass(ModuleToModule):

    def __init__(self, old_module_class, new_module_class, **kwargs):
        super().__init__(new_module_class, **kwargs)
        self.old_module_class = old_module_class

    def apply(self, model: GraphModule) -> GraphModule:
        old_new_module_dict = {}
        for old_module in model.modules():
            # check for equality, not inheritance
            if type(old_module) == self.old_module_class:
                # init the new module based on the old one
                new_module = self._init_new_module(old_module)
                # register modules pair to be replaced
                old_new_module_dict[old_module] = new_module
        # replace all pairs registered
        for old_module, new_module in old_new_module_dict.items():
            self._replace_old_module(model, old_module, new_module)
        return model


class CallableToModule(GraphTransform, ABC):

    def __init__(self, old_callable, new_module_class, **kwargs):
        super().__init__()
        self.old_callable = old_callable
        self.new_module_class = new_module_class
        self.new_module_kwargs = kwargs

    @abstractmethod
    def match_node(self, node: Node):
        pass

    def split_kwargs(self, node: Node):
        new_module_keys = signature_keys(self.new_module_class)
        node_kwargs = dict(node.kwargs)
        node_kwargs_keys = list(node_kwargs.keys())
        module_kwargs = {k: node_kwargs.pop(k) for k in node_kwargs_keys if k in new_module_keys}
        return node_kwargs, module_kwargs

    def move_node_args_to_kwargs(self, node: Node):
        """Move non Node args to kwargs, as long as we can resolve the fn signature somehow"""
        fn = node.target
        if fn in _TORCH_TESTING_DICT:
            fn = _TORCH_TESTING_DICT[fn]
        try:
            fn_kwargs = getcallargs(fn, *node.args, **node.kwargs)
            fn_args = []
            for k, a in list(fn_kwargs.items()):
                if isinstance(a, Node):
                    fn_args.append(fn_kwargs.pop(k))
                else:
                    break
            node.args = tuple(fn_args)
            node.kwargs = immutable_dict(fn_kwargs)
        except TypeError:
            pass

    def rewrite_node(self, node: Node, graph_model: GraphModule):
        module_name = node.name
        assert module_name not in dict(graph_model.named_modules()).keys()
        self.move_node_args_to_kwargs(node)
        node_kwargs, module_kwargs = self.split_kwargs(node)
        module = self.new_module_class(**module_kwargs, **self.new_module_kwargs)
        node.target = module_name
        node.op = 'call_module'
        node.kwargs = immutable_dict(node_kwargs)
        set_module(graph_model, module, module_name)

    def apply(self, graph_model: GraphModule) -> GraphModule:
        for node in graph_model.graph.nodes:
            if self.match_node(node):
                self.rewrite_node(node, graph_model)
        graph_model.recompile()
        graph_model.graph.lint()
        return graph_model


class FnToModule(CallableToModule):

    def match_node(self, node: Node) -> bool:
        return node.op == 'call_function' and node.target is self.old_callable

    def move_node_args_to_kwargs(self, node: Node):
        super().move_node_args_to_kwargs(node)
        # Moving to stateful modules, we remove the 'training' argument if it is passed to the
        # functional version of the layer since it is not needed anymore
        kwargs = dict(node.kwargs)
        if 'training' in kwargs:
            del kwargs['training']
        node.kwargs = immutable_dict(kwargs)


class MethodToModule(CallableToModule):

    def match_node(self, node: Node) -> bool:
        return node.op == 'call_method' and node.target is self.old_callable

    def move_node_args_to_kwargs(self, node: Node):
        if 'self' in node.kwargs:
            node_kwargs = dict(node.kwargs)
            slf = node_kwargs.pop('self')
            node.kwargs = immutable_dict(node_kwargs)
            node.args = tuple([slf] + list(node.args))



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from torch.nn import Module
import torch.nn as nn

from brevitas.export.inference.handler import FloatInferencetHandler
from brevitas.export.inference.handler import FloatWeightInferencetHandler
from brevitas.export.inference.handler import IntInferencetHandler
from brevitas.export.inference.handler import IntWeightInferencetHandler
from brevitas.export.manager import _set_proxy_export_handler
from brevitas.export.manager import _set_proxy_export_mode
from brevitas.export.manager import _set_recurrent_layer_export_handler
from brevitas.export.manager import _set_recurrent_layer_export_mode
from brevitas.export.manager import BaseManager
from brevitas.graph.calibrate import disable_return_quant_tensor
from brevitas.graph.calibrate import restore_return_quant_tensor


def _override_caching_mode(m: nn.Module, attr: str, enabled: bool, metadata_only: bool = True):
    cache_var = 'cache_inference_quant_' + attr
    cache_var_metadata_only = cache_var + '_metadata_only'
    if hasattr(m, cache_var):
        setattr(m, cache_var, enabled)
        setattr(m, cache_var_metadata_only, metadata_only)


def _override_bias_caching_mode(m: nn.Module, enabled: bool, metadata_only: bool = True):
    _override_caching_mode(m, 'bias', enabled, metadata_only)


def _override_act_caching_mode(m: nn.Module, enabled: bool, metadata_only: bool = True):
    _override_caching_mode(m, 'act', enabled, metadata_only)


def _override_weight_caching_mode(m: nn.Module, enabled: bool, metadata_only: bool = False):
    _override_caching_mode(m, 'weight', enabled, metadata_only)


class quant_inference_mode:

    def __init__(self, model, cache_quant_weight=False, enabled=True):
        self.model = model
        self.enabled = enabled
        self.cache_quant_weight = cache_quant_weight
        self.export_manager = InferenceManager
        self.hook_list = []
        self.return_quant_tensor_state = dict()

    def __enter__(self):
        if self.enabled:
            # Register the hook and store it in the list so that it can be removed by the hook itself when called
            handle = self.model.register_forward_hook(self.hook)
            self.hook_list.append(handle)

            # Enable bias for everything. Optionally, store the fully fake-quantized weights
            self.model.apply(
                lambda m: _override_bias_caching_mode(m, enabled=True, metadata_only=True))
            self.model.apply(lambda m: _override_act_caching_mode(m, enabled=True))
            self.model.apply(
                lambda m: _override_weight_caching_mode(
                    m, enabled=True, metadata_only=not self.cache_quant_weight))

    def __exit__(self, type, value, traceback):
        # Disable all caching
        # deactivate export mode
        # restore return quant tensor
        self.model.apply(
            lambda m: _override_bias_caching_mode(m, enabled=False, metadata_only=False))
        self.model.apply(
            lambda m: _override_act_caching_mode(m, enabled=False, metadata_only=False))
        if self.cache_quant_weight:
            self.model.apply(
                lambda m: _override_weight_caching_mode(m, enabled=False, metadata_only=False))
        InferenceManager.set_export_mode(self.model, enabled=False)
        restore_return_quant_tensor(self.model, self.return_quant_tensor_state)

    def hook(self, module, inp, out):
        # After one forward pass with caching enabled, we can:
        # - Set the model in export mode
        # - Attach export handlers
        # - Disable return quant tensor since all quant metadata is cached
        assert len(self.hook_list) == 1
        self.hook_list[0].remove()
        self.model.apply(InferenceManager.set_export_handler)
        InferenceManager.set_export_mode(self.model, enabled=True)
        self.return_quant_tensor_state = disable_return_quant_tensor(self.model)


# Inheritance from BaseManager is not techincally needed
class InferenceManager(BaseManager):
    handlers = [
        IntInferencetHandler,
        FloatInferencetHandler,
        IntWeightInferencetHandler,
        FloatWeightInferencetHandler]

    @classmethod
    def set_export_mode(cls, model: Module, enabled: bool):
        _set_proxy_export_mode(model, enabled)
        _set_recurrent_layer_export_mode(model, enabled)

    @classmethod
    def set_export_handler(cls, module: Module):
        _set_proxy_export_handler(cls, module)
        _set_recurrent_layer_export_handler(cls, module)



"""
Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
"""

from typing import Callable

from torch import Tensor
import torch.nn as nn

from brevitas.core.zero_point import _ScaleShiftZeroPoint
from brevitas.function.ops_ste import abs_binary_sign_grad


# TODO: restore JIT compatibility
class RuntimeDynamicStatsScaling(nn.Module):

    def __init__(
            self,
            scaling_stats_impl: nn.Module,
            dynamic_scaling_broadcastable_fn: Callable,
            scaling_stats_input_view_shape_impl: nn.Module) -> None:
        super(RuntimeDynamicStatsScaling, self).__init__()
        self.scaling_stats_input_view_shape_impl = scaling_stats_input_view_shape_impl
        self.stats_impl = scaling_stats_impl
        self.dynamic_scaling_broadcastable_fn = dynamic_scaling_broadcastable_fn

    def forward(self, x, threshold) -> Tensor:
        shape = x.shape
        x = self.scaling_stats_input_view_shape_impl(x)
        x = self.stats_impl(x) / threshold

        x = self.dynamic_scaling_broadcastable_fn(x, shape)
        return x


# TODO: restore JIT compatibility
class RuntimeDynamicStatsZeroPoint(nn.Module):

    def __init__(
            self,
            zero_point_stats_impl: nn.Module,
            int_quant: nn.Module,
            quantize_zero_point: bool,
            dynamic_scaling_broadcastable_fn: Callable,
            zero_point_stats_input_view_shape_impl: nn.Module) -> None:
        super(RuntimeDynamicStatsZeroPoint, self).__init__()
        self.zero_point_stats_input_view_shape_impl = zero_point_stats_input_view_shape_impl
        self.zero_point_stats_impl = zero_point_stats_impl
        self.dynamic_scaling_broadcastable_fn = dynamic_scaling_broadcastable_fn
        self.scale_shift_zero_point = _ScaleShiftZeroPoint(int_quant, quantize_zero_point)

    def forward(self, x, scale, bit_width) -> Tensor:
        shape = x.shape
        x = self.zero_point_stats_input_view_shape_impl(x)
        x = self.zero_point_stats_impl(x)
        x = self.dynamic_scaling_broadcastable_fn(x, shape)
        x = abs_binary_sign_grad(x)
        x = self.scale_shift_zero_point(x, scale, bit_width)
        return x



from typing import List, NamedTuple, Optional

from torch import Tensor


# Base class for all QuantTensor.
# Only assumptions made by these methods are:
# - `self` is a NamedTuple with a `_fields` attribute
# - `self` has a `value` attribute
class QuantTensor:

    def detach_(self):
        for field in self._fields:
            getattr(self, field).detach_()

    def detach(self):
        qt_type = type(self)
        values = []
        for field in self._fields:
            value = getattr(self, field)
            if isinstance(value, Tensor):
                value = value.detach()
            values.append(value)
        return qt_type(*values)

    def contiguous(self):
        qt_type = type(self)
        values = []
        for field in self._fields:
            value = getattr(self, field)
            if isinstance(value, Tensor):
                value = value.contiguous()
            values.append(value)
        return qt_type(*values)

    def set(self, **kwargs):
        return self._replace(**kwargs)

    @property
    def shape(self):
        return self.value.shape

    def dim(self):
        return self.value.dim()

    def add(self, other):
        return self + other

    def to(self, *args, **kwargs):
        qt_type = type(self)
        values = []
        for field in self._fields:
            value = getattr(self, field)
            if isinstance(value, Tensor):
                value = value.to(*args, **kwargs)
            values.append(value)
        return qt_type(*values)

    def cuda(self, *args, **kwargs):
        qt_type = type(self)
        values = []
        for field in self._fields:
            value = getattr(self, field)
            if isinstance(value, Tensor):
                value = value.cuda(*args, **kwargs)
            values.append(value)
        return qt_type(*values)

    def cpu(self, *args, **kwargs):
        qt_type = type(self)
        values = []
        for field in self._fields:
            value = getattr(self, field)
            if isinstance(value, Tensor):
                value = value.cpu(*args, **kwargs)
            values.append(value)
        return qt_type(*values)

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __pos__(self):
        return self

    def size(self, *args, **kwargs):
        return self.value.size(*args, **kwargs)


class IntQuantTensorBase(NamedTuple):
    value: Tensor
    scale: Tensor
    zero_point: Tensor
    bit_width: Tensor
    signed_t: Tensor
    training_t: Tensor


class FloatQuantTensorBase(NamedTuple):
    value: Tensor
    scale: Tensor
    zero_point: Tensor
    exponent_bit_width: Tensor
    mantissa_bit_width: Tensor
    exponent_bias: Tensor
    saturating_t: Tensor
    inf_values: List[str]
    nan_values: List[str]
    signed_t: Tensor
    training_t: Tensor


class GroupwiseFloatQuantTensorBase(NamedTuple):
    value_: Tensor
    scale_: Tensor
    zero_point_: Tensor
    group_size: Tensor
    group_dim: Tensor
    exponent_bit_width: Tensor
    mantissa_bit_width: Tensor
    exponent_bias: Tensor
    saturating_t: Tensor
    inf_values: List[str]
    nan_values: List[str]
    signed_t: Tensor
    training_t: Tensor


class GroupwisIntQuantTensorBase(NamedTuple):
    value_: Tensor
    scale_: Tensor
    zero_point_: Tensor
    group_size: Tensor
    group_dim: Tensor
    bit_width: Tensor
    signed_t: Tensor
    training_t: Tensor


def _unpack_quant_tensor(input_data):
    if isinstance(input_data, QuantTensor):
        return input_data.value
    elif isinstance(input_data, tuple):
        return tuple([_unpack_quant_tensor(v) for v in input_data])
    elif isinstance(input_data, list):
        return [_unpack_quant_tensor(v) for v in input_data]
    elif isinstance(input_data, dict):
        return {k: _unpack_quant_tensor(v) for k, v in input_data.items()}
    else:
        return input_data



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from abc import abstractmethod
from typing import Any, Optional, Tuple, Union

import torch
from torch import nn
from torch import Tensor
from torch.nn import Identity
from typing_extensions import Protocol
from typing_extensions import runtime_checkable

import brevitas
from brevitas import is_dynamo_compiling
from brevitas.quant_tensor import IntQuantTensor
from brevitas.quant_tensor import QuantTensor
from brevitas.utils.quant_utils import _CachedIO

from .quant_proxy import QuantProxyFromInjector
from .quant_proxy import QuantProxyProtocol

__all__ = [
    'ActQuantProxyProtocol',
    'AccQuantProxyProtocol',
    'ActQuantProxyFromInjector',
    'TruncQuantProxyFromInjector',
    'ClampQuantProxyFromInjector']


def _is_passthrough_act(quant_injector):
    if 'act_impl' not in quant_injector:
        return True
    elif quant_injector.act_impl is None:
        return True
    elif 'passthrough_act' in quant_injector:
        return quant_injector.passthrough_act
    else:
        return False


def _is_act_enabled(act_impl, tensor_quant):
    if act_impl is None:
        return False
    # avoid enabling HardTanh when clamping from quantization is already enabled
    elif isinstance(act_impl, nn.Hardtanh) and tensor_quant is not None:
        return False
    else:
        return True


@runtime_checkable
class ActQuantProxyProtocol(QuantProxyProtocol, Protocol):

    def forward(self, x: Union[Tensor, QuantTensor]) -> QuantTensor:
        ...


@runtime_checkable
class AccQuantProxyProtocol(QuantProxyProtocol, Protocol):

    def forward(self, x: QuantTensor) -> QuantTensor:
        ...


class _TensorQuantDisabledIdentity(brevitas.jit.ScriptModule):

    def __init__(self, module_to_wrap=None):
        super(_TensorQuantDisabledIdentity, self).__init__()

    @brevitas.jit.script_method
    def forward(self,
                x: Tensor) -> Tuple[Tensor, Optional[Tensor], Optional[Tensor], Optional[Tensor]]:
        return (x, None, None, None)


class FusedActivationQuantProxy(brevitas.jit.ScriptModule):

    def __init__(self, activation_impl, tensor_quant):
        super(FusedActivationQuantProxy, self).__init__()
        self.activation_impl = activation_impl
        self.tensor_quant = tensor_quant

    @brevitas.jit.script_method
    def forward(self, x):
        x = self.activation_impl(x)
        return self.tensor_quant(x)


class ActQuantProxyFromInjectorBase(QuantProxyFromInjector, ActQuantProxyProtocol, ABC):

    def __init__(self, quant_layer, quant_injector):
        QuantProxyFromInjector.__init__(self, quant_layer, quant_injector)
        ActQuantProxyProtocol.__init__(self)
        self.is_passthrough_act = _is_passthrough_act(quant_injector)
        self._cached_act = None
        self.cache_inference_quant_act = False
        self.cache_quant_io_metadata_only = True
        self.cache_class = None

    @property
    def input_view_impl(self):
        if self.fused_activation_quant_proxy.tensor_quant is not None and not isinstance(
                self.fused_activation_quant_proxy.tensor_quant, _TensorQuantDisabledIdentity):
            return self.fused_activation_quant_proxy.tensor_quant.int_quant.input_view_impl
        else:
            return Identity()

    def internal_forward(self, force_eval):
        current_status = self.training
        if force_eval:
            self.eval()
        out = self.__call__(self._zero_hw_sentinel())
        self.train(current_status)
        return out

    def retrieve_attribute(self, attribute, force_eval):
        if self._cached_act is not None:
            return getattr(self._cached_act, attribute)
        elif self.is_quant_enabled:
            out = self.internal_forward(force_eval)
            return getattr(out, attribute)
        elif self._cached_act is None:
            return None

    def apply_input_view(self, x):
        return self.input_view_impl(x)

    @property
    def is_quant_enabled(self):
        return self._is_quant_enabled and not self.disable_quant

    @is_quant_enabled.setter
    def is_quant_enabled(self, is_quant_enabled):
        self._is_quant_enabled = is_quant_enabled

    def init_tensor_quant(self):
        tensor_quant = self.quant_injector.tensor_quant
        if 'act_impl' in self.quant_injector:
            act_impl = self.quant_injector.act_impl
        else:
            act_impl = None
        is_act_enabled = _is_act_enabled(act_impl, tensor_quant)
        is_quant_enabled = tensor_quant is not None
        self.is_quant_enabled = is_quant_enabled
        if is_act_enabled and is_quant_enabled:
            self.fused_activation_quant_proxy = FusedActivationQuantProxy(act_impl, tensor_quant)
        elif is_act_enabled and not is_quant_enabled:
            self.fused_activation_quant_proxy = FusedActivationQuantProxy(
                act_impl, _TensorQuantDisabledIdentity())
        elif not is_act_enabled and is_quant_enabled:
            self.fused_activation_quant_proxy = FusedActivationQuantProxy(Identity(), tensor_quant)
        else:
            self.fused_activation_quant_proxy = None

    @abstractmethod
    def create_quant_tensor(
            self,
            qt_args: Union[torch.Tensor, Tuple[Any]],
            x: Optional[QuantTensor] = None) -> QuantTensor:
        # Supports the following:
        # - qt_args as tuple of Tensors and bools = standard quant activations
        # - qt_args as Tensor and x as QuantTensor = passthrough activation
        # In both cases, the output is a QuantTensor
        raise NotImplementedError

    def forward(self, x: Union[Tensor, QuantTensor]) -> Union[Tensor, QuantTensor]:
        # If fused activation quant proxy is not enabled, return the input
        if self.fused_activation_quant_proxy is None:
            return x

        y = x
        if isinstance(y, QuantTensor):
            y = y.value

        if self.export_mode:
            y = self.fused_activation_quant_proxy.activation_impl(y)
            y = self.export_handler(y)
        elif not self.is_quant_enabled:
            # A tuple helps later with control flows
            # The second None value is used later
            # If quant is not enabled, we still apply input_view in the case of groupwise + padding
            y = self.apply_input_view(self.fused_activation_quant_proxy.activation_impl(y))
            y = (y, None)
        else:
            y = self.fused_activation_quant_proxy(y)
        # If y is an empty QuantTensor, we need to check if this is a passthrough proxy,
        # otherwise return a simple Tensor

        if is_dynamo_compiling():
            out = y[0]
        else:
            # If the second value (i.e., scale) is None, then quant is disabled
            if y[1] is not None:
                out = self.create_quant_tensor(y)
            elif self.is_passthrough_act and isinstance(x, QuantTensor):
                # preserve scale/zp/bit/sign even without output quant
                y = y[0]
                out = self.create_quant_tensor(y, x=x)
            else:
                out = y[0]

        if not self.training and self.cache_inference_quant_act and isinstance(out, QuantTensor):
            cached_out = self.cache_class(out.detach(), self.cache_quant_io_metadata_only)
            self._cached_act = cached_out
        return out


class ActQuantProxyFromInjector(ActQuantProxyFromInjectorBase):

    def __init__(self, quant_layer, quant_injector):
        super().__init__(quant_layer, quant_injector)
        self.cache_class = _CachedIO

    def scale(self, force_eval=True):
        return self.retrieve_attribute('scale', force_eval)

    def zero_point(self, force_eval=True):
        return self.retrieve_attribute('zero_point', force_eval)

    def bit_width(self, force_eval=True):
        return self.retrieve_attribute('bit_width', force_eval)

    def create_quant_tensor(
            self,
            qt_args: Union[Tensor, Tuple[Any]],
            x: Optional[IntQuantTensor] = None) -> IntQuantTensor:

        if x is None:
            out = IntQuantTensor(*qt_args, self.is_signed, self.training)
        else:
            out = IntQuantTensor(
                qt_args, x.scale, x.zero_point, x.bit_width, x.signed, self.training)
        return out


class DynamicActQuantProxyFromInjector(ActQuantProxyFromInjector):

    def scale(self, force_eval=True):
        raise RuntimeError("Scale for Dynamic Act Quant is input-dependant")

    def zero_point(self, force_eval=True):
        raise RuntimeError("Zero point for Dynamic Act Quant is input-dependant")


class ClampQuantProxyFromInjector(QuantProxyFromInjector, AccQuantProxyProtocol):

    def forward(self, x: IntQuantTensor) -> Union[Tensor, IntQuantTensor]:
        if self.is_quant_enabled:
            out_tuple = self.tensor_quant(x.value, x.scale, x.bit_width)
            out_value, out_scale, out_zp, out_bit_width = out_tuple
            return IntQuantTensor(
                out_value, out_scale, out_zp, out_bit_width, self.is_signed, self.training)
        return x


class TruncQuantProxyFromInjector(QuantProxyFromInjector, AccQuantProxyProtocol):

    def bit_width(self):
        if not self.is_quant_enabled:
            return None
        zhs = self._zero_hw_sentinel()
        # Signed might or might not be defined. We just care about retrieving the bitwidth
        empty_imp = IntQuantTensor(zhs, zhs, zhs, zhs, signed=True, training=self.training)
        bit_width = self.__call__(empty_imp).bit_width
        return bit_width

    def forward(self, x: IntQuantTensor) -> Union[Tensor, IntQuantTensor]:
        if self.is_quant_enabled:
            if self.export_mode:
                out_tuple = self.export_handler(
                    x.value, x.scale, x.zero_point, x.bit_width, x.signed)
            else:
                out_tuple = self.tensor_quant(x.value, x.scale, x.zero_point, x.bit_width)
            out_value, out_scale, out_zp, out_bit_width = out_tuple
            return IntQuantTensor(
                out_value, out_scale, out_zp, out_bit_width, x.signed, self.training)
        else:
            return x

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        super(TruncQuantProxyFromInjector, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        # for retrocompatibility with when it wasn't removed and it was called differently
        zhs = 'zero_hw_sentinel'
        zhs_key = prefix + zhs
        zhs_old_prefix_key = '.'.join(prefix.split('.')[:-2]) + '.accumulator_quant.' + zhs
        if zhs in unexpected_keys:
            unexpected_keys.remove(zhs_key)
        if zhs_old_prefix_key in unexpected_keys:
            unexpected_keys.remove(zhs_old_prefix_key)



"""
Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
Copyright (c) 2011-2013 NYU                      (Clement Farabet)
Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. Neither the names of Xilinx, Facebook, Deepmind Technologies, NYU,
   NEC Laboratories America and IDIAP Research Institute nor the names
   of its contributors may be used to endorse or promote products derived
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

Forked from PyTorch 2.0.1 with import edits to point to backport
"""

import traceback
from typing import Any, Dict, NamedTuple, Optional, Tuple

import torch

from brevitas.backport.fx._compatibility import compatibility
from brevitas.backport.fx.node import map_aggregate
from brevitas.backport.fx.node import Node

__all__ = ['TensorMetadata', 'ShapeProp']


@compatibility(is_backward_compatible=True)
class TensorMetadata(NamedTuple):
    # TensorMetadata is a structure containing pertinent information
    # about a tensor within a PyTorch program.

    # General Tensor metadata
    shape: torch.Size
    dtype: torch.dtype
    requires_grad: bool
    stride: Tuple[int, ...]
    memory_format: Optional[torch.memory_format]

    # Quantization metadata
    is_quantized: bool
    qparams: Dict[str, Any]


def _extract_tensor_metadata(result: torch.Tensor) -> TensorMetadata:
    """
    Extract a TensorMetadata NamedTuple describing `result`.
    """
    shape = result.shape
    dtype = result.dtype
    requires_grad = result.requires_grad
    stride = result.stride()

    memory_formats = {
        torch.contiguous_format,
        torch.channels_last,
        torch.channels_last_3d,}

    memory_format = None

    for query_format in memory_formats:
        if result.is_contiguous(memory_format=query_format):
            memory_format = query_format
            break

    is_quantized = result.is_quantized
    qparams: Dict[str, Any] = {}
    if is_quantized:
        qscheme = result.qscheme()
        qparams["qscheme"] = qscheme
        if qscheme in {torch.per_tensor_affine, torch.per_tensor_symmetric}:
            qparams["scale"] = result.q_scale()  # type: ignore[assignment]
            qparams["zero_point"] = result.q_zero_point()  # type: ignore[assignment]
        elif qscheme in {torch.per_channel_affine,
                         torch.per_channel_affine_float_qparams,
                         torch.per_channel_symmetric}:
            # In this branch, scale and zero_point are expected to be tensors,
            # we store the values as immutable_list in TensorMetadata for
            # easier serialization downstream
            qparams["scale"] = result.q_per_channel_scales().tolist()  # type: ignore[assignment]
            qparams["zero_point"] = result.q_per_channel_zero_points().tolist(
            )  # type: ignore[assignment]
            qparams["axis"] = result.q_per_channel_axis()  # type: ignore[assignment]

    return TensorMetadata(shape, dtype, requires_grad, stride, memory_format, is_quantized, qparams)


@compatibility(is_backward_compatible=True)
class ShapeProp(torch.fx.Interpreter):
    """
    Execute an FX graph Node-by-Node and
    record the shape and type of the result
    into the corresponding node.

    Example:
         In this example, we record the shape
         and data type of a module given
         an example input ``torch.randn(50, D_in)``.
         We print the name, shape and dtype of each node.

        class TwoLayerNet(torch.nn.Module):
            def __init__(self, D_in, H, D_out):
                super().__init__()
                self.linear1 = torch.nn.Linear(D_in, H)
                self.linear2 = torch.nn.Linear(H, D_out)
            def forward(self, x):
                h_relu = self.linear1(x).clamp(min=0)
                y_pred = self.linear2(h_relu)
                return y_pred
        N, D_in, H, D_out = 64, 1000, 100, 10
        x = torch.randn(N, D_in)
        y = torch.randn(N, D_out)
        model = TwoLayerNet(D_in, H, D_out)
        gm = torch.fx.symbolic_trace(model)
        sample_input = torch.randn(50, D_in)
        ShapeProp(gm).propagate(sample_input)

        for node in gm.graph.nodes:
            print(node.name, node.meta['tensor_meta'].dtype,
                node.meta['tensor_meta'].shape)

        The output of this code is:

        x torch.float32 torch.Size([50, 1000])
        linear1 torch.float32 torch.Size([50, 100])
        clamp_1 torch.float32 torch.Size([50, 100])
        linear2 torch.float32 torch.Size([50, 10])
        output torch.float32 torch.Size([50, 10])

    Args:
         module (GraphModule): The module to be executed
         fake_mode (FakeTensorMode): A fake mode for copying the gm

    """

    def __init__(self, gm, fake_mode=None):
        super().__init__(gm)
        if fake_mode is not None:
            from torch._dynamo.utils import deepcopy_to_fake_tensor

            # Note:
            # We need fake execution cause the inputs are fake, however, we cannot fakify the module
            # - because we need to write to the tensor_meta of the real module. So we fakify to
            # produce a result (L131 below), to extract tensor meta, and then keep going.
            #
            # If we were to fakify, we would write to the wrong node, and then downstream fusion
            # would be missing the tensor_meta.
            #
            # See torch/_inductor/overrides.py for where this is called upstream of fusion.
            self.fake_module = deepcopy_to_fake_tensor(self.module, fake_mode)
            self.fake_mode = fake_mode
        else:
            self.fake_module = None
            self.fake_mode = None

        self.real_module = self.module

    def run_node(self, n: Node) -> Any:
        try:
            if self.fake_module is not None:
                # Hacky swap. Alternatively, we could do this with overriding
                # call_module and get_attr.
                self.module = self.fake_module
            try:
                if self.fake_mode is not None:
                    with self.fake_mode:
                        result = super().run_node(n)
                else:
                    result = super().run_node(n)
            finally:
                self.module = self.real_module
        except Exception as e:
            traceback.print_exc()
            raise RuntimeError(
                f"ShapeProp error for: node={n.format_node()} with "
                f"meta={n.meta}") from e

        found_tensor = False

        def extract_tensor_meta(obj):
            if isinstance(obj, torch.Tensor):
                nonlocal found_tensor
                found_tensor = True
                return _extract_tensor_metadata(obj)
            else:
                return obj

        meta = map_aggregate(result, extract_tensor_meta)
        if found_tensor:
            n.meta['tensor_meta'] = meta

        n.meta['type'] = type(result)
        return result

    def propagate(self, *args):
        """
        Run `module` via interpretation and return the result and
        record the shape and type of each node.

        Args:
            *args (Tensor): the sample input.

        Returns:
            Any: The value returned from executing the Module
        """
        return super().run(*args)



from inspect import signature

import torch

from brevitas.nn.quant_mha import QuantMultiheadAttention

INPUT_NAMES = ['input', 'inp', 'query', 'x', 'hidden_states']


class EqualizedModule(torch.nn.Module):

    def __init__(self, scale_module, layer) -> None:
        super().__init__()
        self.scale = scale_module
        self.layer = layer

    def forward(self, *args, **kwargs):
        # Convert args + kwargs + defaults into kwargs
        bound_arguments = signature(self.layer.forward).bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        kwargs = bound_arguments.arguments

        possible_input_kwargs = INPUT_NAMES
        input_kwarg = [x for x in kwargs.keys() if x in possible_input_kwargs][0]
        x = kwargs[input_kwarg]
        out = x
        if 'key' in kwargs:
            if kwargs['key'].data_ptr() != out.data_ptr():
                raise ValueError(
                    "Cross MHA is not supported for activation equalization."
                    "Replace kwargs with positional args to avoid this exception.")
        out = self.scale(out)

        kwargs[input_kwarg] = out
        # QuantMultiheadAttention is not a subclass of MultiheadAttention
        # We need to preserve the correctness of the forward even after
        # quantization has been applied
        if isinstance(self.layer, (torch.nn.MultiheadAttention, QuantMultiheadAttention)):
            kwargs['key'] = out
            kwargs['value'] = out
        # We convert everything to args so that hooks can work correctly
        out = self.layer(*kwargs.values())
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from torch import Tensor

from brevitas.core.scaling.pre_scaling import AccumulatorAwareParameterPreScaling
from brevitas.function import abs_binary_sign_grad

device = 'cuda' if torch.cuda.is_available() else 'cpu'


def calc_average_psnr(ref_images: Tensor, gen_images: Tensor, eps: float = 1e-10) -> Tensor:
    assert ref_images.shape == gen_images.shape, "Input tensor shapes need to match."
    dist = (ref_images - gen_images).square().mean(axis=(1, 2, 3))  # assuming NCHW
    psnr = 10. * torch.log10(1. / dist.clamp_min(eps))
    return psnr.mean()


def train_for_epoch(trainloader, model, criterion, optimizer, reg_weight: float = 1e-3):
    model.train()

    tot_loss, reg_penalty = 0., 0.

    def acc_reg_penalty(module: AccumulatorAwareParameterPreScaling, inp, output):
        """Accumulate the regularization penalty across constrained layers"""
        nonlocal reg_penalty
        (weights, input_bit_width, input_is_signed) = inp
        s = module.scaling_impl(weights)  # s
        g = abs_binary_sign_grad(module.restrict_clamp_scaling(module.value))  # g
        T = module.calc_max_l1_norm(input_bit_width, input_is_signed)  # T / s
        cur_penalty = torch.relu(g - (T * s)).sum()
        reg_penalty += cur_penalty
        return output

    # Register a forward hook to accumulate the regularization penalty
    hook_fns = list()
    for mod in model.modules():
        if isinstance(mod, AccumulatorAwareParameterPreScaling):
            hook = mod.register_forward_hook(acc_reg_penalty)
            hook_fns.append(hook)

    for _, (images, targets) in enumerate(trainloader):
        optimizer.zero_grad()
        images = images.to(device)
        targets = targets.to(device)
        outputs = model(images)
        task_loss: Tensor = criterion(outputs, targets)
        loss = task_loss + (reg_weight * reg_penalty)
        loss.backward()
        optimizer.step()
        reg_penalty = 0.  # reset the accumulated regularization penalty
        tot_loss += task_loss.item() * images.size(0)

    # Remove the registered forward hooks before exiting
    for hook in hook_fns:
        hook.remove()

    avg_loss = tot_loss / len(trainloader.dataset)
    return avg_loss


@torch.no_grad()
def evaluate_avg_psnr(testloader, model):
    model.eval()
    tot_psnr = 0.
    for _, (images, target) in enumerate(testloader):
        images = images.to(device)
        target = target.to(device)
        output = model(images)
        tot_psnr += calc_average_psnr(output, target).item() * images.size(0)
    avg_psnr = tot_psnr / len(testloader.dataset)
    return avg_psnr



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from .clamp import ClampMin
from .clamp import FloatClamp
from .clamp import ScalarClamp
from .clamp import TensorClamp
from .learned_round import LearnedRoundSte
from .misc import Identity
from .misc import InplaceLogTwo
from .misc import LogTwo
from .misc import PowerOfTwo
from .ops_ste import CeilSte
from .ops_ste import DPURoundSte
from .ops_ste import FloorSte
from .ops_ste import InplaceTensorClampSte
from .ops_ste import RoundSte
from .ops_ste import RoundToZeroSte
from .ops_ste import ScalarClampMinSte
from .ops_ste import TensorClampSte
from .shape import OverBatchOverOutputChannelView
from .shape import OverBatchOverTensorView
from .shape import OverOutputChannelView
from .shape import OverTensorView
from .shape import StatsInputViewShapeImpl
from .stochastic_round import StochasticRoundSte



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, Tuple, Union

from torch import Tensor
from torch.nn import Module

from brevitas.export.manager import _set_proxy_export_handler
from brevitas.export.manager import _set_proxy_export_mode
from brevitas.export.manager import BaseManager
from brevitas.export.manager import ExportContext

from .handler import TorchCDQCastBiasQuantProxyHandler
from .handler import TorchQCDQCastActQuantProxyHandler
from .handler import TorchQCDQCastDecoupledWeightQuantProxyHandler
from .handler import TorchQCDQCastDecoupledWeightQuantWithInputProxyHandler
from .handler import TorchQCDQCastTruncQuantProxyHandler
from .handler import TorchQCDQCastWeightQuantProxyHandler


class TorchQCDQManager(BaseManager):
    target_name = 'torch'

    handlers = [
        TorchQCDQCastWeightQuantProxyHandler,
        TorchQCDQCastDecoupledWeightQuantProxyHandler,
        TorchQCDQCastDecoupledWeightQuantWithInputProxyHandler,
        TorchQCDQCastActQuantProxyHandler,
        TorchCDQCastBiasQuantProxyHandler,
        TorchQCDQCastTruncQuantProxyHandler]

    @classmethod
    def set_export_mode(cls, model: Module, enabled: bool):
        _set_proxy_export_mode(model, enabled)

    @classmethod
    def set_export_handler(cls, module: Module):
        _set_proxy_export_handler(cls, module)

    @classmethod
    def change_weight_export(cls, export_weight_q_node: bool = False):
        for handler in cls.handlers:
            if hasattr(handler, '_export_q_node'):
                handler._export_weight_q_node = export_weight_q_node

    @classmethod
    def export(
            cls,
            module: Module,
            args,
            export_path: Optional[str] = None,
            export_weight_q_node: bool = False):
        cls.change_weight_export(export_weight_q_node=export_weight_q_node)
        with ExportContext(cls):
            traced_module = cls.jit_inference_trace(module, args, export_path)
        return traced_module



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch

from brevitas.config import JIT_ENABLED


def _disabled(fn):
    return fn


if JIT_ENABLED:

    script_method = torch.jit.script_method
    script = torch.jit.script
    ignore = torch.jit.ignore
    ScriptModule = torch.jit.ScriptModule
    Attribute = torch.jit.Attribute

else:

    script_method = _disabled
    script = _disabled
    ignore = _disabled
    ScriptModule = torch.nn.Module
    Attribute = lambda val, type: val



import argparse

import torch
from torch import nn
import torch_mlir

from brevitas.backport.fx._symbolic_trace import wrap
from brevitas.backport.fx.experimental.proxy_tensor import make_fx
from brevitas_examples.llm.llm_quant.export import block_quant_layer_level_manager
from brevitas_examples.llm.llm_quant.export import BlockQuantProxyLevelManager
from brevitas_examples.llm.llm_quant.export import brevitas_layer_export_mode
from brevitas_examples.llm.llm_quant.export import brevitas_proxy_export_mode
from brevitas_examples.llm.llm_quant.export import LinearWeightBlockQuantHandler
from brevitas_examples.llm.llm_quant.export import replace_call_fn_target
from brevitas_examples.llm.llm_quant.mlir_custom_mm import brevitas_matmul_rhs_group_quant_library
from brevitas_examples.llm.llm_quant.quantize import quantize_model


# Due a tracing issue this annotation needs to be
# in the same module (== file) from which make_fx is called
# We also can't directly annotate torch.ops.quant.matmul_rhs_group_quant
# and so we trace a placeholder first and then replace it post tracing
@wrap(visible_to_make_fx=True)
def matmul_rhs_group_quant_placeholder(*args, **kwargs):
    return torch.ops.quant.matmul_rhs_group_quant(*args, **kwargs)


class LinearWeightBlockQuantHandlerFwd(LinearWeightBlockQuantHandler):

    def forward(self, x):
        # Due a tracing issue the call to this fn needs to be
        # in the same module (== file) from which make_fx is called
        out = matmul_rhs_group_quant_placeholder(
            x, self.int_weight, self.scale, self.zero_point, self.bit_width, self.group_size)
        if self.bias is not None:
            out = out + self.bias.view(1, -1)
        return out


class Model(nn.Module):

    def __init__(self):
        super().__init__()
        self.layer = nn.Linear(128, 256, bias=True)

    def forward(self, x):
        return self.layer(x)


def quantize_and_export(args):
    # Init model
    model = Model()

    # Run quantization
    quantize_model(
        model,
        dtype=torch.float32,
        weight_quant_type=args.weight_quant_type,
        weight_bit_width=args.weight_bit_width,
        weight_group_size=args.weight_group_size,
        weight_param_method='stats',
        weight_scale_precision='float',
        weight_quant_granularity='per_group',
        quantize_weight_zero_point=False)

    # Run a test forward pass
    model(torch.randn(2, 128))

    # Pick export mode
    if not args.no_custom_packed_export:
        export_context_manager = brevitas_layer_export_mode
        # we generate an export_class since we need to pass in the handler defined above
        export_class = block_quant_layer_level_manager(
            export_handlers=[LinearWeightBlockQuantHandlerFwd])
    else:
        export_context_manager = brevitas_proxy_export_mode
        export_class = BlockQuantProxyLevelManager

    # export with make_fx with support for fx wrap
    with export_context_manager(model, export_class):
        traced_model = make_fx(model)(torch.randn(2, 128))

    # Replace placeholder for custom op with correct call, if any
    replace_call_fn_target(
        traced_model,
        src=matmul_rhs_group_quant_placeholder,
        target=torch.ops.quant.matmul_rhs_group_quant)

    # print the output graph
    print(traced_model.graph)

    torch_mlir.compile(
        traced_model,
        torch.randn(2, 128),
        output_type="torch",
        backend_legal_ops=["quant.matmul_rhs_group_quant"],
        extra_library=brevitas_matmul_rhs_group_quant_library,
        use_tracing=True,
        verbose=False)


def main():
    parser = argparse.ArgumentParser(
        description='Export single linear with weight group quant to torch-mlir.')
    parser.add_argument('--weight-bit-width', type=int, default=8, help='Weight bit width.')
    parser.add_argument(
        '--weight-quant-type',
        type=str,
        default='asym',
        choices=['sym', 'asym'],
        help='Weight quantization type.')
    parser.add_argument(
        '--weight-group-size',
        type=int,
        default=128,
        help='Group size for group weight quantization.')
    parser.add_argument(
        '--no-custom-packed-export',
        action='store_true',
        help='Enable export to a custom mm op with packed weights for int2 and int4.')
    args = parser.parse_args()
    quantize_and_export(args)


if __name__ == "__main__":
    main()



from dependencies import value, this
from brevitas.core.quant.int import RescalingIntQuant
from brevitas.core.restrict_val import QuantRestrictValue
from brevitas.core.stats.stats_wrapper import SCALAR_SHAPE
from brevitas.inject.enum import ScalingPerOutputType
from brevitas.quant.scaled_int import Int8WeightPerTensorFloat
import brevitas.nn as qnn
import torch


class QuantScalingInt(Int8WeightPerTensorFloat):
    bit_width=8
    module = (this<<1).module
    tracked_parameter_list = (this<<1).tracked_parameter_list
    upstream_scaling = (this<<1).scaling_per_output_type
    rescaling_int_quant = RescalingIntQuant

    @value
    def scaling_shape(scaling_per_output,scaling_per_output_channel_shape, expanded_groupwise_shape, group_dim, upstream_scaling):
        if scaling_per_output == ScalingPerOutputType.TENSOR:
            scaling = SCALAR_SHAPE
        elif scaling_per_output == ScalingPerOutputType.CHANNEL:
            scaling = scaling_per_output_channel_shape
        elif scaling_per_output == ScalingPerOutputType.GROUP:
            # Scaling shape is like expanded_groupwise_shape but has 1 in position group_dim + 1
            assert expanded_groupwise_shape is not None, "Per Group scaling not correctly configured"
            assert group_dim is not None, "Per Group scaling not correctly configured"
            size = list(expanded_groupwise_shape)
            size[group_dim + 1] = 1
            scaling = tuple(size)
        
        # When quantizing scale of groupwise, there will be one extra dim compared to the normal case
        if upstream_scaling == ScalingPerOutputType.GROUP:
            scaling = list(scaling)
            scaling.insert(-1, 1)
            scaling = tuple(scaling)
        return scaling
    

class QuantScaleInt8WeightPerTensorFloat(Int8WeightPerTensorFloat):
    scaling_int_quant = QuantScalingInt
    restrict_scaling_impl = QuantRestrictValue
    scaling_per_output_type = ScalingPerOutputType.CHANNEL
    group_size=32 

    @value
    def restrict_value_float_to_int_impl():
        return this.scaling_int_quant.rescaling_int_quant

linear = qnn.QuantLinear(64, 768, weight_quant=QuantScaleInt8WeightPerTensorFloat)
o = linear(torch.randn(1, 64))
print("----------------")
print(torch.max(torch.abs(linear.weight - linear.quant_weight().value)))





# class QuantScalingInt(SolveStatsReduceDimFromEnum,
#                         SolveWeightScalingStatsInputDimsFromModule,
#                         SolveScalingStatsInputViewShapeImplFromEnum,
#                         SolveScalingStatsOpFromEnum,
#                         SolveBitWidthImplFromEnum,
#                         SolveTensorQuantFloatToIntImplFromEnum,
#                         SolveRestrictScalingImplFromEnum,
#                         SolveIntScalingImplFromEnum,
#                         SolveParameterScalingImplFromEnum,
#                         SolveParameterTensorClampImplFromEnum,
#                         SolveParameterScalingInitFromEnum,
#                         SolveParameterScalingShape,
#                         SolveWeightScalingPerOutputChannelShapeFromModule,
#                         SolveWeightTensorQuantFromEnum,
#                         SolveDtypeDeviceFromTrackedParameterList,
#                         SolveInputViewImpl):
#     scaling_int_quant = None
#     rescaling_int_quant = RescalingIntQuant
#     quant_type = QuantType.INT
#     bit_width_impl_type = BitWidthImplType.CONST
#     float_to_int_impl_type = FloatToIntImplType.ROUND
#     narrow_range = True
#     signed = True
#     zero_point_impl = ZeroZeroPoint
#     scaling_impl_type = ScalingImplType.STATS
#     scaling_stats_op = StatsOp.MAX
#     scaling_min_val = 1e-10
#     scaling_per_output_type = ScalingPerOutputType.CHANNEL
#     restrict_scaling_type = RestrictValueType.FP
#     bit_width = 8
#     module = (this<<1).module
#     tracked_parameter_list = (this<<1).tracked_parameter_list
#     upstream_scaling = (this<<1).scaling_per_output_type

#     @value
#     def scaling_shape(scaling_per_output,scaling_per_output_channel_shape, expanded_groupwise_shape, group_dim, upstream_scaling):
#         if scaling_per_output == ScalingPerOutputType.TENSOR:
#             scaling = SCALAR_SHAPE
#         elif scaling_per_output == ScalingPerOutputType.CHANNEL:
#             scaling = scaling_per_output_channel_shape
#         elif scaling_per_output == ScalingPerOutputType.GROUP:
#             # Scaling shape is like expanded_groupwise_shape but has 1 in position group_dim + 1
#             assert expanded_groupwise_shape is not None, "Per Group scaling not correctly configured"
#             assert group_dim is not None, "Per Group scaling not correctly configured"
#             size = list(expanded_groupwise_shape)
#             size[group_dim + 1] = 1
#             scaling = tuple(size)
#         if upstream_scaling == ScalingPerOutputType.GROUP:
#             scaling = list(scaling)
#             scaling.insert(-1, 1)
#             scaling = tuple(scaling)
#         return scaling
    


# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Tuple, Union

import torch
from torch.nn import Module
from torch.nn import Parameter

import brevitas
import brevitas.config as config
from brevitas.core.function_wrapper import Identity
from brevitas.core.restrict_val import _RestrictClampValue
from brevitas.core.restrict_val import FloatRestrictValue
from brevitas.core.stats import _ParameterListStats
from brevitas.core.stats import _RuntimeStats
from brevitas.core.stats import DEFAULT_MOMENTUM
from brevitas.core.utils import ParameterWrapper
from brevitas.core.utils import StatelessBuffer
from brevitas.function.ops_ste import abs_binary_sign_grad


class StatsFromParameterScaling(brevitas.jit.ScriptModule):

    def __init__(
            self,
            scaling_stats_impl: Module,
            scaling_stats_input_view_shape_impl: Module,
            scaling_stats_input_concat_dim: int,
            tracked_parameter_list: List[torch.nn.Parameter],
            scaling_shape: Tuple[int, ...],
            restrict_scaling_impl: Module = FloatRestrictValue(),
            affine_rescaling: bool = False,
            affine_shift_scale: bool = False,
            scaling_min_val: Optional[float] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(StatsFromParameterScaling, self).__init__()
        self.parameter_list_stats = _ParameterListStats(
            scaling_stats_impl,
            scaling_shape,
            scaling_stats_input_view_shape_impl,
            scaling_stats_input_concat_dim,
            tracked_parameter_list)
        self.stats_scaling_impl = _StatsScaling(
            restrict_scaling_impl,
            scaling_shape,
            scaling_min_val,
            affine_rescaling,
            affine_shift_scale,
            dtype,
            device)

    @brevitas.jit.script_method
    def forward(
            self, ignored: torch.Tensor, threshold: Optional[torch.Tensor] = None) -> torch.Tensor:
        stats = self.parameter_list_stats()
        if threshold is None:
            threshold = torch.ones(1).type_as(stats)
        return self.stats_scaling_impl(stats, threshold)


class _StatsScaling(brevitas.jit.ScriptModule):

    def __init__(
            self,
            restrict_scaling_impl: Module,
            scaling_shape: Tuple[int, ...],
            scaling_min_val: Optional[float],
            affine_rescaling: bool,
            affine_shift_scale: bool,
            dtype: Optional[torch.dtype],
            device: Optional[torch.device]) -> None:
        super(_StatsScaling, self).__init__()
        if affine_shift_scale and not affine_rescaling:
            raise RuntimeError(
                "Disabling shifting of the scale requires to enable affine rescaling first.")
        if affine_rescaling:
            self.affine_rescaling = _AffineRescaling(
                scaling_shape, affine_shift_scale, dtype, device)
        else:
            self.affine_rescaling = Identity()
        self.restrict_clamp_scaling = _RestrictClampValue(scaling_min_val, restrict_scaling_impl)
        self.restrict_scaling_pre = restrict_scaling_impl.restrict_init_module()
        self.restrict_scaling_impl = restrict_scaling_impl

    @brevitas.jit.script_method
    def forward(
            self, stats: torch.Tensor, threshold: Optional[torch.Tensor] = None) -> torch.Tensor:
        if threshold is None:
            threshold = torch.ones(1).type_as(stats)
        threshold = self.restrict_scaling_pre(threshold)
        stats = self.restrict_scaling_pre(stats)
        stats = self.restrict_scaling_impl.combine_scale_threshold(stats, threshold)
        stats = self.affine_rescaling(stats)
        stats = self.restrict_clamp_scaling(stats)
        return stats


class RuntimeStatsScaling(brevitas.jit.ScriptModule):

    def __init__(
            self,
            scaling_stats_impl: Module,
            scaling_stats_input_view_shape_impl: Module,
            scaling_shape: Tuple[int, ...],
            affine_rescaling: bool = False,
            affine_shift_scale: bool = False,
            restrict_scaling_impl: Module = FloatRestrictValue(),
            scaling_stats_momentum: float = DEFAULT_MOMENTUM,
            scaling_min_val: Optional[float] = None,
            dtype: Optional[torch.dtype] = None,
            device: Optional[torch.device] = None) -> None:
        super(RuntimeStatsScaling, self).__init__()

        self.runtime_stats = _RuntimeStats(
            scaling_stats_impl,
            scaling_shape,
            scaling_stats_input_view_shape_impl,
            scaling_stats_momentum,
            dtype,
            device)
        self.stats_scaling_impl = _StatsScaling(
            restrict_scaling_impl,
            scaling_shape,
            scaling_min_val,
            affine_rescaling,
            affine_shift_scale,
            dtype,
            device)

    @brevitas.jit.script_method
    def forward(self, x: torch.Tensor, threshold: Optional[torch.Tensor] = None) -> torch.Tensor:
        stats = self.runtime_stats(x)
        return self.stats_scaling_impl(stats, threshold)


class _AffineRescaling(brevitas.jit.ScriptModule):

    def __init__(
            self,
            scaling_shape,
            shift_scale,
            dtype: Optional[torch.dtype],
            device: Optional[torch.device]):
        super(_AffineRescaling, self).__init__()
        self.affine_weight = Parameter(torch.ones(scaling_shape, dtype=dtype, device=device))
        if shift_scale:
            self.affine_bias = ParameterWrapper(
                torch.zeros(scaling_shape, dtype=dtype, device=device))
        else:
            self.affine_bias = StatelessBuffer(torch.tensor(0., dtype=dtype, device=device))

    @brevitas.jit.script_method
    def forward(self, x):
        out = x * self.affine_weight + self.affine_bias()
        out = abs_binary_sign_grad(out)
        return out

    def _load_from_state_dict(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys,
            error_msgs):
        super(_AffineRescaling, self)._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)
        affine_weight_key = prefix + 'affine_weight'
        affine_bias_key = prefix + 'affine_bias'
        if config.IGNORE_MISSING_KEYS and affine_weight_key in missing_keys:
            missing_keys.remove(affine_weight_key)
        if config.IGNORE_MISSING_KEYS and affine_bias_key in missing_keys:
            missing_keys.remove(affine_bias_key)


class RuntimeDynamicGroupStatsScaling(brevitas.jit.ScriptModule):

    def __init__(
        self,
        group_size: int,
        group_dim: int,
        input_view_impl: Module,
        scaling_stats_impl: Module,
        scaling_min_val: Optional[float],
        restrict_scaling_impl: Module = FloatRestrictValue()) -> None:
        super(RuntimeDynamicGroupStatsScaling, self).__init__()
        self.group_size = group_size
        self.group_dim = group_dim
        self.scaling_stats_impl = scaling_stats_impl
        self.scaling_min_val = scaling_min_val
        self.input_view_impl = input_view_impl
        self.restrict_clamp_scaling = _RestrictClampValue(scaling_min_val, restrict_scaling_impl)

    @brevitas.jit.script_method
    def forward(
            self,
            stats_input: torch.Tensor,
            threshold: Optional[torch.Tensor] = None) -> torch.Tensor:
        if threshold is None:
            threshold = torch.ones(1).type_as(stats_input)
        stats_input_reshaped = self.input_view_impl(stats_input)
        out = self.scaling_stats_impl(stats_input_reshaped) / threshold
        # Scaling min val
        out = self.restrict_clamp_scaling(out)
        return out



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import argparse
import json
import os
import pprint
import random

import numpy as np
import torch

from brevitas_examples.super_resolution.models import get_model_by_name
from brevitas_examples.super_resolution.utils import device
from brevitas_examples.super_resolution.utils import evaluate_accumulator_bit_widths
from brevitas_examples.super_resolution.utils import evaluate_avg_psnr
from brevitas_examples.super_resolution.utils import export
from brevitas_examples.super_resolution.utils import get_bsd300_dataloaders

random_seed = 123456

random.seed(random_seed)
np.random.seed(random_seed)
torch.manual_seed(random_seed)

desc = """Evaluating single-image super resolution models on the BSD300 dataset.

Example:
>> python eval_model.py --data_root=data --model=quant_espcn_x2_w8a8_a2q_16b --use_pretrained --export_to_qonnx
"""

parser = argparse.ArgumentParser(description='PyTorch BSD300 Validation')
parser.add_argument('--data_root', help='Path to folder containing BSD300 val folder')
parser.add_argument('--model_path', default=None, help='Path to PyTorch checkpoint. Default = None')
parser.add_argument(
    '--save_path',
    type=str,
    default='outputs/',
    help='Save path for exported model. Default = outputs/')
parser.add_argument(
    '--model',
    type=str,
    default='quant_espcn_x2_w8a8_base',
    help='Name of the model configuration. Default = quant_espcn_x2_w8a8_base')
parser.add_argument(
    '--workers', type=int, default=0, help='Number of data loading workers. Default = 0')
parser.add_argument('--batch_size', type=int, default=16, help='Minibatch size. Default = 16')
parser.add_argument(
    '--crop_size', type=int, default=512, help='The size to crop the image. Default = 512')
parser.add_argument('--use_pretrained', action='store_true', default=False)
parser.add_argument('--eval_acc_bw', action='store_true', default=False)
parser.add_argument('--save_model_io', action='store_true', default=False)
parser.add_argument('--export_to_qonnx', action='store_true', default=False)
parser.add_argument('--export_to_qcdq_onnx', action='store_true', default=False)
parser.add_argument('--export_to_qcdq_torch', action='store_true', default=False)


def main():
    args = parser.parse_args()

    # initialize model, dataset, and training environment
    model = get_model_by_name(args.model, args.use_pretrained)
    if args.model_path is not None:
        assert not args.use_pretrained, "Already loading pretrained model."
        model.load_state_dict(torch.load(args.model_path, map_location='cpu'))
    model = model.to(device)
    _, testloader = get_bsd300_dataloaders(
        args.data_root,
        num_workers=args.workers,
        batch_size=args.batch_size,
        upscale_factor=model.upscale_factor,
        crop_size=args.crop_size,
        download=True)

    test_psnr = evaluate_avg_psnr(testloader, model)
    print(f"[{args.model}] test_psnr={test_psnr:.2f}")

    os.makedirs(args.save_path, exist_ok=True)

    # evaluate accumulator bit widths
    if args.eval_acc_bw:
        inp = testloader.dataset[0][0].unsqueeze(0).to(device)
        stats_dict = {
            'acc_bit_widths': evaluate_accumulator_bit_widths(model, inp), 'test_psnr': test_psnr}
        with open(f"{args.save_path}/stats.json", "w") as outfile:
            json.dump(stats_dict, outfile, indent=4)
        pretty_stats_dict = pprint.pformat(stats_dict, sort_dicts=False)
        print(pretty_stats_dict)

    # save and export model
    export(model, testloader, args)


if __name__ == '__main__':
    main()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import Optional, Tuple, Type, Union

from packaging import version
import torch
from torch import Tensor
from torch.nn import ConvTranspose1d
from torch.nn import ConvTranspose2d
from torch.nn import ConvTranspose3d
from torch.nn.functional import conv_transpose1d
from torch.nn.functional import conv_transpose2d
from torch.nn.functional import conv_transpose3d

from brevitas import torch_version
from brevitas.function.ops import max_int
from brevitas.function.ops_ste import ceil_ste
from brevitas.inject.defaults import Int8WeightPerTensorFloat
from brevitas.quant_tensor import QuantTensor

from .quant_layer import ActQuantType
from .quant_layer import BiasQuantType
from .quant_layer import QuantWeightBiasInputOutputLayer as QuantWBIOL
from .quant_layer import WeightQuantType

__all__ = ['QuantConvTranspose1d', 'QuantConvTranspose2d', 'QuantConvTranspose3d']


class QuantConvTranspose1d(QuantWBIOL, ConvTranspose1d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: int,
            stride: int = 1,
            padding: int = 0,
            output_padding: int = 0,
            dilation: int = 1,
            groups: int = 1,
            padding_mode: str = 'zeros',
            bias: Optional[bool] = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            **kwargs) -> None:
        ConvTranspose1d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
            bias=bias,
            device=device,
            dtype=dtype)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self._output_size = None

    @property
    def per_elem_ops(self):
        raise NotImplementedError

    @property
    def output_channel_dim(self) -> int:
        return 1

    @property
    def channelwise_separable(self) -> bool:
        raise self.groups == self.out_channels

    def forward(self,
                input: Union[Tensor, QuantTensor],
                output_size=None) -> Union[Tensor, QuantTensor]:
        self._output_size = output_size  # cache the value temporarily
        return self.forward_impl(input)

    def compute_output_padding(self, inp, output_size):
        if torch_version >= version.parse('1.12'):
            return self._output_padding(
                inp, output_size, self.stride, self.padding, self.kernel_size, num_spatial_dims=1)
        else:
            return self._output_padding(
                inp, output_size, self.stride, self.padding, self.kernel_size)

    def conv_transpose1d_zeros_pad(
            self, x: Tensor, weight: Tensor, bias: Optional[Tensor], output_padding):
        out = conv_transpose1d(
            x,
            weight,
            bias,
            stride=self.stride,
            padding=self.padding,
            output_padding=output_padding,
            groups=self.groups,
            dilation=self.dilation)
        return out

    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.padding_mode == 'zeros':
            output_padding = self.compute_output_padding(x, self._output_size)
            self._output_size = None  # set it back to None after consuming it
            out = self.conv_transpose1d_zeros_pad(x, quant_weight, quant_bias, output_padding)
            return out
        else:
            raise NotImplementedError(f"Padding mode {self.padding_mode} not supported.")


class QuantConvTranspose2d(QuantWBIOL, ConvTranspose2d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple[int]],
            stride: Union[int, Tuple[int]] = 1,
            padding: Union[int, Tuple[int]] = 0,
            output_padding: Union[int, Tuple[int]] = 0,
            dilation: Union[int, Tuple[int]] = 1,
            groups: int = 1,
            padding_mode: str = 'zeros',
            bias: Optional[bool] = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            **kwargs) -> None:
        ConvTranspose2d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
            bias=bias,
            device=device,
            dtype=dtype)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self._output_size = None

    @property
    def per_elem_ops(self):
        raise NotImplementedError

    @property
    def output_channel_dim(self) -> int:
        return 1

    @property
    def channelwise_separable(self) -> bool:
        raise self.groups == self.out_channels

    def forward(self,
                input: Union[Tensor, QuantTensor],
                output_size=None) -> Union[Tensor, QuantTensor]:
        self._output_size = output_size  # cache the value temporarily
        return self.forward_impl(input)

    def compute_output_padding(self, inp, output_size):
        if torch_version >= version.parse('1.12'):
            return self._output_padding(
                inp, output_size, self.stride, self.padding, self.kernel_size, num_spatial_dims=2)
        else:
            return self._output_padding(
                inp, output_size, self.stride, self.padding, self.kernel_size)

    def conv_transpose2d_zeros_pad(
            self, x: Tensor, weight: Tensor, bias: Optional[Tensor], output_padding):
        out = conv_transpose2d(
            x,
            weight,
            bias,
            stride=self.stride,
            padding=self.padding,
            output_padding=output_padding,
            groups=self.groups,
            dilation=self.dilation)
        return out

    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.padding_mode == 'zeros':
            output_padding = self.compute_output_padding(x, self._output_size)
            self._output_size = None  # set it back to None after consuming it
            out = self.conv_transpose2d_zeros_pad(x, quant_weight, quant_bias, output_padding)
            return out
        else:
            raise NotImplementedError(f"Padding mode {self.padding_mode} not supported.")


class QuantConvTranspose3d(QuantWBIOL, ConvTranspose3d):

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple[int]],
            stride: Union[int, Tuple[int]] = 1,
            padding: Union[int, Tuple[int]] = 0,
            output_padding: Union[int, Tuple[int]] = 0,
            dilation: Union[int, Tuple[int]] = 1,
            groups: int = 1,
            padding_mode: str = 'zeros',
            bias: bool = True,
            weight_quant: Optional[WeightQuantType] = Int8WeightPerTensorFloat,
            bias_quant: Optional[BiasQuantType] = None,
            input_quant: Optional[ActQuantType] = None,
            output_quant: Optional[ActQuantType] = None,
            return_quant_tensor: bool = False,
            device: Optional[torch.device] = None,
            dtype: Optional[torch.dtype] = None,
            **kwargs) -> None:
        ConvTranspose3d.__init__(
            self,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            padding_mode=padding_mode,
            bias=bias,
            device=device,
            dtype=dtype)
        QuantWBIOL.__init__(
            self,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
            input_quant=input_quant,
            output_quant=output_quant,
            return_quant_tensor=return_quant_tensor,
            **kwargs)
        self._output_size = None

    @property
    def per_elem_ops(self):
        raise NotImplementedError

    @property
    def output_channel_dim(self) -> int:
        return 1

    @property
    def channelwise_separable(self) -> bool:
        raise self.groups == self.out_channels

    def forward(self,
                input: Union[Tensor, QuantTensor],
                output_size=None) -> Union[Tensor, QuantTensor]:
        self._output_size = output_size  # cache the value temporarily
        return self.forward_impl(input)

    def compute_output_padding(self, inp, output_size):
        if torch_version >= version.parse('1.12'):
            return self._output_padding(
                inp, output_size, self.stride, self.padding, self.kernel_size, num_spatial_dims=3)
        else:
            return self._output_padding(
                inp, output_size, self.stride, self.padding, self.kernel_size)

    def conv_transpose3d_zeros_pad(
            self, x: Tensor, weight: Tensor, bias: Optional[Tensor], output_padding):
        out = conv_transpose3d(
            x,
            weight,
            bias,
            stride=self.stride,
            padding=self.padding,
            output_padding=output_padding,
            groups=self.groups,
            dilation=self.dilation)
        return out

    def inner_forward_impl(self, x: Tensor, quant_weight: Tensor, quant_bias: Optional[Tensor]):
        if self.padding_mode == 'zeros':
            output_padding = self.compute_output_padding(x, self._output_size)
            self._output_size = None  # set it back to None after consuming it
            out = self.conv_transpose3d_zeros_pad(x, quant_weight, quant_bias, output_padding)
            return out
        else:
            raise NotImplementedError(f"Padding mode {self.padding_mode} not supported.")



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import copy
from typing import List, Optional, Tuple

import torch
from torch.nn import Sequential

import brevitas
from brevitas.function.ops_ste import floor_ste


class TupleSequential(Sequential):

    def output(self, mod, input):
        if isinstance(input, tuple):
            return mod(*input)
        else:
            return mod(input)

    def forward(self, *input):
        modules = list(self._modules.values())
        out = self.output(modules[0], input)
        for mod in modules[1:]:
            out = self.output(mod, out)
        return out


class KwargsForwardHook(torch.nn.Module):

    def __init__(self, module, hook_fn):
        super().__init__()
        self.module = module
        self.hook_fn = hook_fn

    def forward(self, *args, **kwargs):
        out = self.module(*args, **kwargs)
        args = args + (out,)
        self.hook_fn(self.module, *args, **kwargs)
        return out


def torch_partial_deepcopy(model):
    """
    Performs a deepcopy of a torch.nn.Module, except for all the parameters that are instead passed by reference
    """
    memo = {}
    for p in model.parameters():
        memo[id(p)] = copy.copy(p)  # Shallow copy of parameters
    model_copy = copy.deepcopy(model, memo)
    return model_copy


def kthvalue(
    x: torch.Tensor,
    k: int,
    dim: Optional[int] = None,
    keepdim: bool = False,
    out: Optional[Tuple[torch.Tensor, torch.LongTensor]] = None
) -> Tuple[torch.Tensor, torch.LongTensor]:
    # As of torch 2.1, there is no kthvalue implementation:
    # - In CPU for float16
    # - In GPU for bfloat16
    # In these cases we cast to float32 and then go back to the original dtype
    dtype = x.dtype
    device = str(x.device)

    # We do not support out as buffer for the output, since we cannot control its dtype
    if out is not None:
        raise RuntimeError("out argument for kthvalue not supported")

    if (dtype == torch.float16 and 'cpu' in device) or \
        (dtype == torch.bfloat16 and 'cuda' in device):
        x = x.type(torch.float32)

    # PyTorch specify None as default for `dim` but it breaks if we specifically pass None
    if dim is not None:
        x, indices = torch.kthvalue(x, k, dim=dim, keepdim=keepdim)
    else:
        x, indices = torch.kthvalue(x, k, keepdim=keepdim)

    if x.dtype != dtype:
        x = x.type(dtype)
    return (x, indices)


def compute_channel_view_shape(tensor: torch.Tensor, channel_dim: int):
    broadcast_shape = [1] * len(tensor.size())
    broadcast_shape[channel_dim] = -1
    return tuple(broadcast_shape)


@brevitas.jit.script
def float_internal_scale(
        x: torch.Tensor,
        mantissa_bit_width: torch.Tensor,
        fp_internal_scale_min: torch.Tensor,
        eps: float) -> torch.Tensor:

    internal_scale = floor_ste(torch.log2(torch.abs(x) + eps)) - mantissa_bit_width
    internal_scale = torch.clamp_min(internal_scale, fp_internal_scale_min)
    internal_scale = torch.exp2(internal_scale)
    return internal_scale


@brevitas.jit.ignore
def padding(x: torch.Tensor, group_size: int, group_dim: int) -> List[int]:
    # Given a tensor X, compute the padding aloing group_dim so that groupwise shaping is possible
    padding = [0, 0] * len(x.shape)
    size = x.shape
    if size[group_dim] % group_size != 0:
        padding[2 * group_dim] = group_size - size[group_dim] % group_size
    padding = list(reversed(padding))
    return padding
